#!/usr/bin/env python
# -*- coding: utf-8 -*-
# (c) Copyright IBM Corp. 2010, 2021. All Rights Reserved.

import logging
import os

from resilient import constants as res_constants
from resilient import get_config_file
from resilient import helpers as res_helpers
from resilient import parse_parameters

import resilient_circuits.keyring_arguments as keyring_arguments
from resilient_circuits import constants
from resilient_circuits.helpers import validate_configs
from resilient_circuits.validate_configs import VALIDATE_DICT


class AppArgumentParser(keyring_arguments.ArgumentParser):
    """Helper to parse command line arguments."""
    DEFAULT_APP_SECTION = "resilient"
    DEFAULT_STOMP_PORT = 65001
    DEFAULT_COMPONENTS_DIR = ''
    DEFAULT_LOG_LEVEL = 'INFO'
    DEFAULT_LOG_FILE = 'app.log'
    DEFAULT_LOG_MAX_BYTES = 10000000
    DEFAULT_LOG_BACKUP_COUNT = 10
    DEFAULT_NO_PROMPT_PASS = "False"
    DEFAULT_STOMP_TIMEOUT = 60
    DEFAULT_STOMP_MAX_RETRIES = 3
    DEFAULT_MAX_CONNECTION_RETRIES = res_constants.APP_CONFIG_MAX_CONNECTION_RETRIES_DEFAULT
    DEFAULT_NUM_WORKERS = 25
    DEFAULT_APP_EXCEPTION = False
    DEFAULT_HEARTBEAT_TIMEOUT_THRESHOLD = None

    def __init__(self, config_file=None):

        self._setup_temp_logger()
        config_file = config_file or get_config_file()
        super(AppArgumentParser, self).__init__(config_file=config_file)

        default_components_dir = self.getopt(self.DEFAULT_APP_SECTION, "componentsdir") or self.DEFAULT_COMPONENTS_DIR
        default_noload = self.getopt(self.DEFAULT_APP_SECTION, "noload") or ""
        default_log_dir = self.getopt(self.DEFAULT_APP_SECTION, "logdir") or constants.APP_LOG_DIR
        default_log_level = self.getopt(self.DEFAULT_APP_SECTION, "loglevel") or self.DEFAULT_LOG_LEVEL
        default_log_file = self.getopt(self.DEFAULT_APP_SECTION, "logfile") or self.DEFAULT_LOG_FILE
        default_log_max_bytes = self.getopt(self.DEFAULT_APP_SECTION, constants.APP_CONFIG_LOG_MAX_BYTES) or self.DEFAULT_LOG_MAX_BYTES
        default_log_backup_count = self.getopt(self.DEFAULT_APP_SECTION, constants.APP_CONFIG_LOG_BACKUP_COUNT) or self.DEFAULT_LOG_BACKUP_COUNT

        # STOMP port is usually 65001
        default_stomp_port = self.getopt(self.DEFAULT_APP_SECTION, "stomp_port") or self.DEFAULT_STOMP_PORT
        # For some environments the STOMP TLS certificate will be different from the REST API cert
        default_stomp_cafile = self.getopt(self.DEFAULT_APP_SECTION, "stomp_cafile") or None

        default_stomp_url = self.getopt(self.DEFAULT_APP_SECTION, "stomp_host") or self.getopt(self.DEFAULT_APP_SECTION, "host")
        default_stomp_timeout = self.getopt(self.DEFAULT_APP_SECTION, "stomp_timeout") or self.DEFAULT_STOMP_TIMEOUT
        default_stomp_max_retries = self.getopt(self.DEFAULT_APP_SECTION, "stomp_max_retries") or self.DEFAULT_STOMP_MAX_RETRIES
        default_max_connection_retries = self.getopt(self.DEFAULT_APP_SECTION, res_constants.APP_CONFIG_MAX_CONNECTION_RETRIES) or self.DEFAULT_MAX_CONNECTION_RETRIES

        default_no_prompt_password = self.getopt(self.DEFAULT_APP_SECTION,
                                                 "no_prompt_password") or self.DEFAULT_NO_PROMPT_PASS
        default_no_prompt_password = self._is_true(default_no_prompt_password)

        default_test_actions = self._is_true(self.getopt(self.DEFAULT_APP_SECTION,
                                                         "test_actions")) or False
        default_test_host = self.getopt(self.DEFAULT_APP_SECTION, "test_host") or None
        default_test_port = self.getopt(self.DEFAULT_APP_SECTION, "test_port") or None
        default_log_responses = self.getopt(self.DEFAULT_APP_SECTION,
                                            "log_http_responses") or ""
        default_resource_prefix = self.getopt(self.DEFAULT_APP_SECTION, "resource_prefix") or None
        default_num_workers = self.getopt(self.DEFAULT_APP_SECTION, "num_workers") or self.DEFAULT_NUM_WORKERS
        default_trap_exception = self.getopt(self.DEFAULT_APP_SECTION, constants.APP_CONFIG_TRAP_EXCEPTION) or self.DEFAULT_APP_EXCEPTION
        default_trap_exception = self._is_true(default_trap_exception)

        default_heartbeat_timeout_threshold = self.getopt(self.DEFAULT_APP_SECTION, constants.APP_CONFIG_HEARTBEAT_TIMEOUT_THRESHOLD) or self.DEFAULT_HEARTBEAT_TIMEOUT_THRESHOLD

        self._unset_temp_logger()

        self.add_argument("--stomp-host",
                          type=str,
                          default=default_stomp_url,
                          help="Resilient server STOMP host url")
        self.add_argument("--stomp-port",
                          type=int,
                          default=default_stomp_port,
                          help="Resilient server STOMP port number")
        self.add_argument("--stomp-cafile",
                          type=str,
                          action="store",
                          default=default_stomp_cafile,
                          help="Resilient server STOMP TLS certificate")
        self.add_argument("--stomp-timeout",
                          type=int,
                          action="store",
                          default=os.environ.get('RESILIENT_STOMP_TIMEOUT', default_stomp_timeout),
                          help="Resilient server STOMP timeout for connections")
        self.add_argument("--stomp-max-retries",
                          type=int,
                          action="store",
                          default=os.environ.get('RESILIENT_STOMP_MAX_RETRIES', default_stomp_max_retries),
                          help="Resilient server STOMP max retries before failing")
        self.add_argument("--max-connection-retries",
                          type=int,
                          action="store",
                          default=os.environ.get('RESILIENT_MAX_CONNECTION_RETRIES') or 0 if os.environ.get("APP_HOST_CONTAINER") else default_max_connection_retries,
                          help="Number of attempts to retry when connecting to SOAR. Use 0 or -1 for unlimited retries. Defaults to 0")
        self.add_argument("--resource-prefix",
                          type=str,
                          action="store",
                          default=os.environ.get('RESOURCE_PREFIX', default_resource_prefix),
                          help="Cloud Pak for Security resource path for host and STOMP URLs")
        self.add_argument("--componentsdir",
                          type=str,
                          default=default_components_dir,
                          help="Circuits components auto-load directory")
        self.add_argument("--noload",
                          type=str,
                          default=default_noload,
                          help="List of components that should not be loaded")
        self.add_argument("--logdir",
                          type=str,
                          default=default_log_dir,
                          help="Directory for log files")
        self.add_argument("--loglevel",
                          type=str,
                          default=default_log_level,
                          help="Log level")
        self.add_argument("--logfile",
                          type=str,
                          default=default_log_file,
                          help="File to log to")
        self.add_argument("--{0}".format(constants.APP_CONFIG_LOG_MAX_BYTES),
                          type=int,
                          default=default_log_max_bytes,
                          help="Maximum bytes per log file")
        self.add_argument("--{0}".format(constants.APP_CONFIG_LOG_BACKUP_COUNT),
                          type=int,
                          default=default_log_backup_count,
                          help="Number of log files to create in rotation")
        self.add_argument("--no-prompt-password",
                          type=bool,
                          default=default_no_prompt_password,
                          help="Never prompt for password on stdin")
        self.add_argument("--test-actions",
                          action="store_true",
                          default=default_test_actions,
                          help="Enable submitting test actions?")
        self.add_argument("--test-host",
                          type=str,
                          action="store",
                          default=default_test_host,
                          help=("For use with --test-actions option. "
                                "Host or IP to bind test server to."))
        self.add_argument("--test-port",
                          type=int,
                          action="store",
                          default=default_test_port,
                          help=("For use with --test-actions option. "
                                "Port to bind test server to."))
        self.add_argument("--log-http-responses",
                          type=str,
                          default=default_log_responses,
                          help=("Log all responses from Resilient "
                                "REST API to this directory"))
        self.add_argument("--num-workers",
                          type=int,
                          default=default_num_workers,
                          help=("Number of FunctionWorkers to use. "
                                "Number of Functions that can run in parallel"))
        self.add_argument("--trap-exception",
                          type=bool,
                          default=default_trap_exception,
                          help=("If set to 'True' a Function's exception will be ignored"))
        self.add_argument("--{0}".format(constants.APP_CONFIG_HEARTBEAT_TIMEOUT_THRESHOLD),
                          type=int,
                          default=default_heartbeat_timeout_threshold,
                          help=("The amount of time in seconds that can occur between HeartbeatTimeouts before exiting"))

    def parse_args(self, args=None, namespace=None, ALLOW_UNRECOGNIZED=False):
        """Parse commandline arguments and construct an opts dictionary"""
        self._setup_temp_logger()
        opts = super(AppArgumentParser, self).parse_args(args, namespace, ALLOW_UNRECOGNIZED)
        if self.config:
            for section in self.config.sections():
                items = dict((item.lower(), self.config.get(section, item)) for item in self.config.options(section))
                opts.update({section: items})

            parse_parameters(opts)

            # Once we have read the app.config and decrypted any protected secrets
            # we must remove the secrets directory
            res_helpers.remove_secrets_dir()

        validate_configs(opts, VALIDATE_DICT)

        # NOTE: Newer retry2 logic requires tries to be -1 (not 0 as before) for unlimited attempts
        if opts.get(res_constants.APP_CONFIG_MAX_CONNECTION_RETRIES, -1) == 0:
            opts[res_constants.APP_CONFIG_MAX_CONNECTION_RETRIES] = -1

        self._unset_temp_logger()
        return opts

    def _setup_temp_logger(self):
        # Temporary logging handler until the real one is created later in app.py
        # return
        self.temp_handler = logging.StreamHandler()
        self.temp_handler.setFormatter(logging.Formatter('%(asctime)s %(levelname)s [%(module)s] %(message)s'))
        self.temp_handler.setLevel(logging.INFO)
        logging.getLogger().addHandler(self.temp_handler)

    def _unset_temp_logger(self):
        # Unset the temporary logger
        # return
        logging.getLogger().removeHandler(self.temp_handler)

    @staticmethod
    def _is_true(value):
        if value:
            return value.lower()[0] in ("1", "t", "y")
        else:
            return False



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# (c) Copyright IBM Corp. 2010, 2021. All Rights Reserved.

import pytest
from mock import patch
from resilient_circuits import helpers
from resilient_circuits.actions_component import Actions
from resilient_circuits.stomp_events import HeartbeatTimeout
from resilient_lib import IntegrationError

from tests import MockInboundAppComponent
from tests import helpers as test_helpers
from tests import mock_constants
from tests.shared_mock_data import mock_paths

resilient_mock = mock_constants.RESILIENT_MOCK
config_data = mock_constants.CONFIG_DATA


def test_inbound_cmp_raises_error_no_app_configs(circuits_app):
    mock_cmp = MockInboundAppComponent(opts=mock_constants.MOCK_OPTS)
    mock_cmp.app_configs = None
    mock_cmp.register(circuits_app.app.component_loader)
    with pytest.raises(IntegrationError, match=r"does not have app_configs defined"):
        test_helpers.call_inbound_app(circuits_app, mock_constants.MOCK_INBOUND_Q_NAME)


@patch("resilient_circuits.actions_component.helpers.get_fn_names", new=lambda x: [])
def test_actions_on_heartbeat_timeout_no_config(circuits_app, fx_simple_client, caplog):

    mock_configs = helpers.get_configs(path_config_file=mock_paths.MOCK_APP_CONFIG)
    mock_configs.pop("heartbeat_timeout_threshold")

    with patch("resilient_circuits.actions_component.ResilientComponent.rest_client") as mock_client:
        mock_client.return_value = fx_simple_client[0]
        mock_actions_cmp = Actions(opts=mock_configs)

        mock_actions_cmp.on_heartbeat_timeout(HeartbeatTimeout(1))
        mock_actions_cmp.on_heartbeat_timeout(HeartbeatTimeout(2))
        mock_actions_cmp.on_heartbeat_timeout(HeartbeatTimeout(3))
        mock_actions_cmp.on_heartbeat_timeout(HeartbeatTimeout(4))
        mock_actions_cmp.on_heartbeat_timeout(HeartbeatTimeout(5))

        assert "Trying to reconnect after STOMP HeartbeatTimeout" in caplog.text
        assert "exiting..." not in caplog.text


@patch("resilient_circuits.actions_component.helpers.get_fn_names", new=lambda x: [])
def test_actions_on_heartbeat_timeout_with_config_set(circuits_app, fx_simple_client, caplog):

    mock_configs = helpers.get_configs(path_config_file=mock_paths.MOCK_APP_CONFIG)

    with patch("resilient_circuits.actions_component.ResilientComponent.rest_client") as mock_client:
        mock_client.return_value = fx_simple_client[0]
        mock_actions_cmp = Actions(opts=mock_configs)

        mock_actions_cmp.on_heartbeat_timeout(HeartbeatTimeout(10))
        assert "Trying to reconnect after STOMP HeartbeatTimeout" in caplog.text

        with pytest.raises(SystemExit) as sys_exit:
            mock_actions_cmp.on_heartbeat_timeout(HeartbeatTimeout(20))

        exit_msg = "'{0}' is set to '{1}s' and '{2}s' have passed since the first HeartbeatTimeout, exiting...".format(
            "heartbeat_timeout_threshold", "5", "10"
        )

        assert exit_msg in caplog.text
        assert sys_exit.type == SystemExit
        assert sys_exit.value.code == 34



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# (c) Copyright IBM Corp. 2010, 2021. All Rights Reserved.

"""Common paths used in tests"""

import os

SHARED_MOCK_DATA_DIR = os.path.dirname(os.path.realpath(__file__))

MOCK_APP_CONFIG = os.path.join(SHARED_MOCK_DATA_DIR, "mock_app_config")
MOCK_COMMENTED_APP_CONFIG = os.path.join(SHARED_MOCK_DATA_DIR, "mock_commented_app_config")
MOCK_COMPONENT = os.path.join(SHARED_MOCK_DATA_DIR, "mock_component.py")
MOCK_IMPORT_DEFINITION = os.path.join(SHARED_MOCK_DATA_DIR, "mock_import_definition.txt")


#!/usr/bin/env python
# -*- coding: utf-8 -*-
# (c) Copyright IBM Corp. 2010, 2021. All Rights Reserved.

import pytest
from resilient_lib import IntegrationError, RequestsCommon
from resilient_circuits import StatusMessage
from tests import mock_constants, AppFunctionMockComponent


resilient_mock = mock_constants.RESILIENT_MOCK
config_data = mock_constants.CONFIG_DATA


def test_basic_instantiation(circuits_app):
    mock_cmp = AppFunctionMockComponent(
        opts=mock_constants.MOCK_OPTS,
        package_name=mock_constants.MOCK_PACKAGE_NAME,
        required_app_configs=mock_constants.MOCK_REQUIRED_APP_CONFIGS)

    assert mock_cmp.PACKAGE_NAME == mock_constants.MOCK_PACKAGE_NAME
    assert mock_cmp.opts == mock_constants.MOCK_OPTS
    assert mock_cmp.required_app_configs == mock_constants.MOCK_REQUIRED_APP_CONFIGS
    assert isinstance(mock_cmp.rc, RequestsCommon)
    assert mock_cmp.app_configs.url == "https://www.mockexample.com"
    assert mock_cmp.options == mock_cmp._app_configs_as_dict


def test_status_message(circuits_app):
    mock_msg = u"Custom message with unicode լ խ ծ կ հ ձ ղ ճ"
    mock_cmp = AppFunctionMockComponent(opts=mock_constants.MOCK_OPTS)
    mock_status_message = mock_cmp.status_message(mock_msg)
    assert isinstance(mock_status_message, StatusMessage)
    assert mock_status_message.text == mock_msg


def test_log_message(circuits_app, caplog):
    mock_msg = u"Custom message with unicode լ խ ծ կ հ ձ ղ ճ"
    mock_cmp = AppFunctionMockComponent(opts=mock_constants.MOCK_OPTS, package_name=mock_constants.MOCK_PACKAGE_NAME)
    mock_cmp.LOG.info(mock_msg)
    assert mock_msg in caplog.text



# -*- coding: utf-8 -*-
# (c) Copyright IBM Corp. 2010, 2019. All Rights Reserved.

"""Circuits component for Action Module subscription and message handling"""

import base64
import json
import logging
import os.path
import ssl
import sys
import traceback
from signal import SIGINT, SIGTERM

if sys.version_info.major < 3:
    from collections import Callable
else:
    from collections.abc import Callable

from circuits import BaseComponent, Worker
from circuits.core.handlers import handler
from circuits.core.manager import ExceptionWrapper
from requests.utils import DEFAULT_CA_BUNDLE_PATH
from six import string_types

import resilient
from resilient import ensure_unicode
from resilient_lib import IntegrationError
import resilient_circuits.actions_test_component as actions_test_component
from resilient_circuits import constants, helpers
from resilient_circuits.action_message import (ActionMessage,
                                               ActionMessageBase,
                                               BaseFunctionError,
                                               FunctionMessage, FunctionResult,
                                               InboundMessage, StatusMessage)
from resilient_circuits.decorators import *  # for back-compatibility, these were previously declared here
from resilient_circuits.rest_helper import (get_resilient_client,
                                            reset_resilient_client)
from resilient_circuits.stomp_component import StompClient
from resilient_circuits.stomp_events import *

LOG = logging.getLogger(__name__)

STOMP_CLIENT_HEARTBEAT = 0          # no heartbeat from client to server
STOMP_SERVER_HEARTBEAT = 15000      # 15-second heartbeat from server to client
RETRY_TIMER_INTERVAL = 60           # Retry failed deliveries every minute
SUBSCRIBE_TO_QUEUES_TIMEOUT = 30    # connect event timeout

# Global idle timer, fires after 10 minutes to reset the REST connection
IDLE_TIMER_INTERVAL = 600
_idle_timer = None

# look for unrecoverable errors
UNRECOVERABLE_ERRORS = ['Already subscribed']

# Errors and Message Destination names will get appended when ran in selftest
SELFTEST_ERRORS = []
SELFTEST_SUBSCRIPTIONS = []


def validate_cert(cert, hostname):
    """Utility wrapper for SSL validation on the STOMP connection"""
    try:
        resilient.match_hostname(cert, hostname)
    except Exception as exc:
        return (False, str(exc))
    return (True, "Success")


class FunctionWorker(Worker):

    channel = "functionworker"

    @handler("signal", channel="*")
    def _on_signal(self, signo, stack):
        """Add a signal handler to the worker processes otherwise they swallow SIGINT, SIGTERM
           (see FallBackSignalHandler in circuits/core/helpers.py)
        """
        if signo in [SIGINT, SIGTERM]:
            LOG.error("Worker interrupted")
            raise SystemExit(0)

    @handler("task", override=True)
    def _on_task(self, f, *args, **kwargs):
        LOG.debug("Task: %s", f)
        result = self.pool.apply_async(f, args, kwargs)
        while not result.ready():
            yield
        try:
            yield result.get()
        except Exception as e:
            str_traceback = traceback.format_exc()
            ignore_exception = False

            if isinstance(self.parent, Actions):
                app_configs = self.parent.opts
                ignore_exception = app_configs.get(constants.APP_CONFIG_TRAP_EXCEPTION, False)

            if ignore_exception:

                err = str(e)

                # Handle strs as unicode if Python 2
                if sys.version_info.major == 2:
                    err = unicode(err, "utf-8")
                    str_traceback = unicode(str_traceback, "utf-8")

                fn_name = constants.DEFAULT_UNKNOWN_STR

                # args being the parameters of the decorator
                if isinstance(args, tuple) and hasattr(args[0], "name"):
                    fn_name = args[0].name

                status_message = u"Error running '{0}'. Config '{1}' set to 'True' so ignoring exception\nERROR:\n{2}".format(fn_name, constants.APP_CONFIG_TRAP_EXCEPTION, err)

                # If the loglevel is DEBUG, the full stacktrace will be added to the StatusMessage
                if logging.getLogger().getEffectiveLevel() == logging.DEBUG:
                    status_message = u"{0}\n{1}".format(status_message, str_traceback)

                log_message = u"{0}\n{1}".format(status_message, str_traceback)

                yield StatusMessage(status_message)

                LOG.warning(log_message)

                yield FunctionResult({"success": False, "reason": err})

            else:
                LOG.error(str_traceback)
                yield ExceptionWrapper(e)


class ResilientComponent(BaseComponent):
    """A Circuits base component with a connection to the Resilient APIs.

       This is the superclass for custom Resilient Action Module components.
       If a component inherits from ResilientComponent, it will automatically be loaded,
       and actions and functions will be dispatched to its `handler` and `function` methods.
    """
    test_mode = False  # True with --test-actions option
    IS_SELFTEST = False

    def __init__(self, opts):
        super(ResilientComponent, self).__init__()
        assert isinstance(opts, dict)

        self.opts = opts
        self._fields = {}
        self._action_fields = {}
        self._functions = {}
        self._function_fields = {}
        self.fn_names = helpers.get_fn_names(self)

        # Get fields, message destinations and functions

        self._get_fields(fn_names=self.fn_names)

        # If this component is a function, self.fn_names will be defined
        if self.fn_names:
            LOG.debug("@function handler names: %s", self.fn_names)

            for fn_name in self.fn_names:
                if not helpers.check_exists(fn_name, self._functions):
                    LOG.warning("Function '{0}' is not defined in this Resilient platform!".format(fn_name))

        else:
            # Check that decorated requirements are met
            callables = ((x, getattr(self, x)) for x in dir(self) if isinstance(getattr(self, x), Callable))
            for name, func in callables:
                if name == "__class__":
                    name = func.__name__
                # Do all the custom fields exist?
                fields = getattr(func, "required_fields", {})
                for (field_name, input_type) in fields.items():
                    try:
                        fielddef = self._fields[field_name]
                    except KeyError:
                        errmsg = ("Field '{0}' (required by '{1}') is not "
                                "defined in this Resilient platform.")
                        raise Exception(errmsg.format(field_name, name))
                    if input_type is not None:
                        if fielddef["input_type"] != input_type:
                            errmsg = ("Field '{0}' (required by '{1}') "
                                    "must be type '{2}'.")
                            raise Exception(errmsg.format(field_name,
                                                        name, input_type))
                # Do all the action fields exist?
                fields = getattr(func, "required_action_fields", {})
                for (field_name, input_type) in fields.items():
                    try:
                        fielddef = self._action_fields[field_name]
                    except KeyError:
                        errmsg = ("Action field '{0}' (required by '{1}') "
                                "is not defined in this Resilient platform.")
                        raise Exception(errmsg.format(field_name, name))
                    if input_type is not None:
                        if fielddef["input_type"] != input_type:
                            errmsg = ("Action field '{0}' (required by "
                                    "'{1}') must be type '{2}'.")
                            raise Exception(errmsg.format(field_name,
                                                        name, input_type))

    def _get_fields(self, fn_names=None):
        """Get Incident and Action fields"""
        client = self.rest_client()
        self._fields = dict((field["name"], field)
                            for field in client.cached_get("/types/incident/fields"))
        self._action_fields = dict((field["name"], field)
                                   for field in client.cached_get("/types/actioninvocation/fields"))

        if fn_names:

            try:

                for fn_name in fn_names:
                    self._functions[fn_name] = client.cached_get("/functions/{0}?handle_format=names".format(fn_name))

                self._function_fields = dict((field["name"], field) for field in client.cached_get("/types/__function/fields"))

            except resilient.SimpleHTTPException:
                # functions are not available, pre-v30 server
                self._functions = None
                self._function_fields = None

    def rest_client(self):
        """
        Return a connected instance of the :class:`resilient.SimpleClient <resilient.co3.SimpleClient>`
        that can be used to access the IBM SOAR REST API.

        .. note::
            An :class:`~resilient_circuits.app_function_component.AppFunctionComponent` inherits a
            ``ResilientComponent`` so will have access to this method

        **Example:**

        .. code-block:: python

            from resilient_circuits import AppFunctionComponent, FunctionResult, app_function

            PACKAGE_NAME = "fn_mock_app"
            FN_NAME = "mock_function"

            class FunctionComponent(AppFunctionComponent):

                def __init__(self, opts):
                    super(FunctionComponent, self).__init__(opts, PACKAGE_NAME)

                @app_function(FN_NAME)
                def _app_function(self, fn_inputs):

                    res_client = self.rest_client()
                    inc = res_client.get("/incidents/{0}".format("1001"))
                    inc_name = inc.get("name")

                    results = {
                        "inc": inc,
                        "inc_name": inc_name
                    }

                    yield FunctionResult(results)

        :return: a connected instance of :class:`resilient.SimpleClient <resilient.co3.SimpleClient>`
        :rtype: :class:`resilient.SimpleClient <resilient.co3.SimpleClient>`

        """
        self.reset_idle_timer()
        return get_resilient_client(self.opts)

    def reset_idle_timer(self):
        """Create an idle-timer that we can use to reset the REST connection"""
        global _idle_timer
        if _idle_timer is None:
            LOG.debug("create idle timer")
            _idle_timer = Timer(IDLE_TIMER_INTERVAL,
                                Event.create("idle_reset"), persist=True)
            _idle_timer.register(self)
        else:
            _idle_timer.reset()

    def get_incident_field(self, fieldname):
        """Get the definition of an incident-field"""
        return self._fields[fieldname]

    def get_incident_fields(self):
        """Get the definitions of all incident-fields, as a list"""
        return self._fields.values()

    def get_field_label(self, fieldname, value_id):
        """Get the label for an incident-field value id"""
        field = self._fields[fieldname]
        for value in field["values"]:
            if value["enabled"]:
                if value["value"] == value_id:
                    return value["label"]
        return value_id

    def get_action_field(self, fieldname):
        """Get the definition of an action-field"""
        return self._action_fields[fieldname]

    def get_action_fields(self):
        """Get the definitions of all action-fields, as a list"""
        return self._action_fields.values()

    def get_action_field_label(self, fieldname, value_id):
        """Get the label for an action-field value id"""
        field = self._action_fields[fieldname]
        for value in field["values"]:
            if value["enabled"]:
                if value["value"] == value_id:
                    return value["label"]
        return ensure_unicode(value_id)  # fallback

    def get_function_field(self, fieldname):
        """Get the definition of a function input field (parameter)"""
        return self._function_fields[fieldname] if self._function_fields else None

    def get_function_fields(self):
        """Get the definitions of all function input-fields (parameters), as a list"""
        return self._function_fields.values() if self._function_fields else None

    def get_function_field_label(self, fieldname, value_id):
        """Get the label for a function input-field value id"""
        if self._function_fields is None:
            return None
        field = self._function_fields[fieldname]
        for value in field["values"]:
            if value["enabled"]:
                if value["value"] == value_id:
                    return value["label"]
        return ensure_unicode(value_id)  # fallback

    @staticmethod
    def get_select_param(value):
        """Function Helper: get the label from a select- or multi-select parameter"""
        if isinstance(value, list):
            return [ResilientComponent.get_select_param(val) for val in value]
        if isinstance(value, dict):
            return value.get("name")
        return value

    @staticmethod
    def get_textarea_param(value):
        """Function Helper: get the content from a textarea-type parameter"""
        if isinstance(value, dict):
            return value.get("content")
        return value

    @handler("reload")
    def reload(self, event, opts):
        """Event handler called when the configuration options have changed."""
        self.opts = opts
        self._get_fields()


class Actions(ResilientComponent):
    """Component that subscribes to Resilient Action Module queues and fires message events"""

    # Whenever a component in the circuit is registered to a channel name "actions.xxxx",
    # this component will subscribe to the corresponding Action Module queue (xxxx)
    # and then fire events for each message that arrives from the queue.
    # After the message is handled, or fails, it acks the message and updates the action status.

    def __init__(self, opts):
        super(Actions, self).__init__(opts)
        self.listeners = dict()
        self._proxy_args = {}

        # messages and acks that failed to send over stomp connection
        self._stomp_ack_delivery_failures = {}
        self._resilient_ack_delivery_failures = {}

        # Read the action definitions, into a dict indexed by id
        # we'll refer to them later when dispatching
        self.reconnect_stomp = True
        self.org_id = None
        self.action_defs = dict()
        self.stomp_component = None
        self.logging_directory = None
        self.subscribe_headers = None
        self._configure_opts(opts)
        self.max_retry_count = int(opts.get("stomp_max_retries")) # default from app.py:DEFAULT_STOMP_MAX_RETRIES
        self.stomp_timeout = int(opts.get("stomp_timeout", SUBSCRIBE_TO_QUEUES_TIMEOUT))
        self.heartbeat_timeouts = []

        timer_internal = int(opts['resilient'].get("stomp_timer_interval", RETRY_TIMER_INTERVAL))

        _retry_timer = Timer(timer_internal, Event.create("retry_failed_deliveries"), persist=True)
        _retry_timer.register(self)

        # Make a worker thread-pool that will run functions
        LOG.debug("num_workers set to %s", opts.get("num_workers"))
        self._num_workers = opts.get("num_workers")
        self._functionworker = FunctionWorker(process=False, channel="functionworker", workers=opts.get("num_workers"))
        self._functionworker.register(self.root)

        if opts.get("test_actions", False):
            # Let user submit test actions from the command line for testing
            LOG.info("Action Tests enabled. Run 'resilient-circuits test' for the interactive action test tool.")
            test_options = {}
            if opts.get("test_port", None):
                test_options["port"] = int(opts["test_port"])
            if self.opts.get("test_host"):
                test_options["host"] = opts["test_host"]
            actions_test_component.ResilientTestActions(self.org_id,
                                                        **test_options).register(self)

    def _configure_opts(self, opts):
        """ Handle settings from configuration """
        # Set options for connecting to Action Module with HTTP Proxy
        self._proxy_args = {}
        proxy_host = opts.get("proxy_host")
        if proxy_host:
            proxy_host = proxy_host.replace("http://", "").replace("https://", "")
            self._proxy_args = {"proxy_host": proxy_host,
                                "proxy_port": opts.get("proxy_port"),
                                "proxy_user": opts.get("proxy_user"),
                                "proxy_password": opts.get("proxy_password")}

        rest_client = self.rest_client()
        self.org_id = rest_client.org_id

        list_action_defs = rest_client.get("/actions")["entities"]
        self.action_defs = dict((int(action["id"]), action) for action in list_action_defs)

        self.subscribe_headers = {"activemq.prefetchSize": opts["stomp_prefetch_limit"]}

    # Public Utility methods

    def action_name(self, action_id):
        """Get the name of an action, from its id"""
        if action_id is None:
            # Unnamed action, probably triggered from a v28 workflow
            LOG.info("Action: _unnamed_")
            return "_unnamed_"

        if self.action_defs.get(action_id):
            defn = self.action_defs[action_id]
        else:
            LOG.warning("Action %s is unknown.", action_id)
            # Refresh the list of action definitions
            list_action_defs = self.rest_client().get("/actions")["entities"]
            self.action_defs = dict((int(action["id"]),
                                     action) for action in list_action_defs)

            defn = self.action_defs.get(action_id, {"name": "_unnamed_"})

        if defn:
            return defn.get("name", "_unnamed_")

    @handler("Connected")
    def on_stomp_connected(self):
        """Client has connected to the STOMP server"""
        LOG.info("STOMP connected.")

        # Once STOMP has connected, we want to empty the heartbeat_timeouts list
        self.heartbeat_timeouts = []

        # Stop retrying the failed deliveries from previous session
        # We'll ack them right away if they are re-delivered
        # TODO: We should age these out of here at some point. 24 hours?
        for key in self._stomp_ack_delivery_failures:
            self._stomp_ack_delivery_failures[key]["from_prev_conn"] = True
        for key in self._resilient_ack_delivery_failures:
            self._resilient_ack_delivery_failures[key]["from_prev_conn"] = True

    @handler("HeartbeatTimeout")
    def on_heartbeat_timeout(self, event):
        """
        Handler for a HeartbeatTimeout event

        Checks to see if the 'heartbeat_timeout_threshold' config is set. Keeps
        track of HBs in a list. Each HB has a ts value. If the difference
        between the first and current HB's ts is greater than the config,
        resilient-circuits exits with an exit code of 34
        """

        heartbeat_timeout_threshold = self.opts.get(constants.APP_CONFIG_HEARTBEAT_TIMEOUT_THRESHOLD)

        # If heartbeat_timeout_threshold config is set
        if heartbeat_timeout_threshold:

            # Add the current HB to the list
            self.heartbeat_timeouts.append(event)

            # If more than 1 HBs are in the list
            if len(self.heartbeat_timeouts) > 1:

                # Remove any invalid timeouts and sort the list
                self.heartbeat_timeouts = helpers.filter_heartbeat_timeout_events(self.heartbeat_timeouts)

                # Get the difference in seconds between the oldest and current timeout
                delta = int(self.heartbeat_timeouts[-1].ts - self.heartbeat_timeouts[0].ts)

                # If the delta is greater than the config, log an error and exit resilient-circuits
                if delta > heartbeat_timeout_threshold:
                    LOG.error("'%s' is set to '%ss' and '%ss' have passed since the first HeartbeatTimeout, exiting...",
                              constants.APP_CONFIG_HEARTBEAT_TIMEOUT_THRESHOLD, heartbeat_timeout_threshold, delta)

                    sys.exit(constants.EXIT_STOMP_HEARTBEAT_TIMEOUT)

        LOG.error("Trying to reconnect after STOMP HeartbeatTimeout")

        # Fire a Disconnect event to Reconnect
        self.fire(Disconnect(flush=False, reconnect=True))

    @handler("OnStompError")
    def on_stomp_error(self, headers, message, error):
        """STOMP produced an error."""
        LOG.error('STOMP listener: Error:\n%s', message or error)

        if helpers.is_this_a_selftest(self):
            SELFTEST_ERRORS.append(message or error)

        for unrecoverable_error in UNRECOVERABLE_ERRORS:
            if (message and unrecoverable_error in str(message)) or \
               (error and unrecoverable_error in str(error)):
                LOG.error("Unrecoverable error. Exiting")
                raise SystemExit(0)

        # Just raise the event for anyone listening
        self.fire(Event("exception", "Actions",
                        headers.get("message"), message))

    @handler("Message")
    def on_stomp_message(self, event, headers, message, queue):
        """STOMP produced a message."""
        # Find the queue name from the subscription id (stomp_listener_xxx)
        msg_id = event.frame.headers.get("message-id")
        if not msg_id:
            LOG.error("Received message with no message id. %s", event.frame.info())
            raise ValueError("Stomp message with no message id received")
        elif msg_id in self._resilient_ack_delivery_failures or msg_id in self._stomp_ack_delivery_failures:
            # This is a message we have already processed but we failed to acknowledge
            # Don't process it again, just acknowledge it
            LOG.info("Skipping reprocess of message %s.  Sending saved ack now.", msg_id)
            failure_info = self._resilient_ack_delivery_failures.get(msg_id)
            if failure_info:
                self.fire(Send(headers={'correlation-id': headers['correlation-id']},
                               body=failure_info["result"]["body"],
                               destination=headers['reply-to'],
                               message_id=msg_id))
                self._resilient_ack_delivery_failures.pop(msg_id)
            failure_info = self._stomp_ack_delivery_failures.get(msg_id)
            if failure_info:
                self.fire(Ack(event.frame))
                self._stomp_ack_delivery_failures.pop(msg_id)

        else:
            subscription = self.stomp_component.get_subscription(event.frame)
            LOG.debug('STOMP listener: message for %s', subscription)
            queue_name = subscription.split(".", 2)[2]
            channel = "actions." + queue_name

            LOG.debug("Got Message: %s", event.frame.info())

            try:
                # Expect the message payload to always be UTF8 JSON.
                # However, it may contain surrogate pairs, and in Python 3 that causes problems:
                # - surrogate pairs are not allowed by the default (strict) utf8 decoder,
                # - if we pass them, it will cause downstream issues, so we should re-encode.
                try:
                    mstr = message.decode('utf-8')
                except UnicodeDecodeError:
                    LOG.debug("Failed utf8 decode, trying surrogate")
                    mstr = message.decode('utf-8', "surrogatepass").encode("utf-16", "surrogatepass").decode("utf-16")

                message = json.loads(mstr)
                # Construct a Circuits event with the message, and fire it on the channel
                if queue and queue[0] == constants.INBOUND_MSG_DEST_PREFIX:
                    channel = u"{0}.{1}".format(constants.INBOUND_MSG_DEST_PREFIX, queue[2])
                    event = InboundMessage(source=self,
                                           queue=queue,
                                           headers=headers,
                                           message=message,
                                           frame=event.frame,
                                           log_dir=self.logging_directory)

                elif message.get("function"):
                    channel = "functions." + message["function"]["name"]
                    event = FunctionMessage(source=self,
                                            headers=headers,
                                            message=message,
                                            frame=event.frame,
                                            log_dir=self.logging_directory)
                else:
                    event = ActionMessage(source=self,
                                          headers=headers,
                                          message=message,
                                          frame=event.frame,
                                          log_dir=self.logging_directory)
                LOG.info("Event: %s Channel: %s", event, channel)

                self.fire(event, channel)
            except Exception as exc:
                LOG.exception(exc)
                if not isinstance(message, dict):
                    LOG.error("DATA:%s", base64.b64encode(message))
                # Normally the event won't be ack'd.  Just report it and carry on.
                if self.ignore_message_failure:
                    # Construct and fire anyway, which will ack the message
                    LOG.warning("This message failure will be ignored...")
                    event = ActionMessage(source=self,
                                          headers=headers,
                                          message=None,
                                          frame=event.frame,
                                          log_dir=self.logging_directory)
                    self.fire(event, channel)

    # Circuits event handlers

    @handler("idle_reset")
    def idle_reset(self, event):
        LOG.debug("Idle reset")
        reset_resilient_client()

    def _setup_stomp(self):
        rest_client = self.rest_client()
        if not rest_client.actions_enabled:
            # Don't create stomp connection b/c action module is not enabled.
            LOG.warning(("Resilient action module not enabled."
                        "No stomp connection attempted."))
            return

        self.resilient_mock = self.opts["resilient_mock"] or False
        if self.resilient_mock:
            # Using mock API, no need to create a real stomp connection
            LOG.warning("Using Mock. No Stomp connection")
            return

        # Gather the stomp_cafile for if specified or fallback to the resilient host. Used for TLS / SSL
        cafile = self.opts.get("stomp_cafile") or self.opts.cafile
        if cafile is not None and cafile.strip().lower() == "false":
            # Explicitly disable TLS certificate validation, if you need to
            cafile = None
            LOG.warning(("Unverified STOMP TLS certificate (cafile=false)"))
        elif cafile is None:
            # Since the REST API (resilient library) uses 'requests', let's use its default certificate bundle
            # instead of the certificates from ssl.get_default_verify_paths().cafile
            cafile = DEFAULT_CA_BUNDLE_PATH
            LOG.debug("STOMP TLS validation with default certificate file: %s", cafile)
        else:
            LOG.debug("STOMP TLS validation with certificate file: %s", cafile)

        try:
            ca_certs = None
            context = ssl.create_default_context(cafile=cafile)
            context.check_hostname = True if cafile else False
            context.verify_mode = ssl.CERT_REQUIRED if cafile else ssl.CERT_NONE
        except AttributeError as err:
            # Likely an older ssl version w/out true ssl context
            LOG.info("Can't create SSL context. Using fallback method")
            context = None
            ca_certs = cafile

        # Gather the stomp_host if specified or fallback to the resilient host if not
        stomp_host = self.opts["resilient"].get("stomp_host", None) or self.opts["host"]
        #
        #   key_id and key_secret are the preferrable one
        #
        if self.opts.get("api_key_id", None) is not None and self.opts.get("api_key_secret", None) is not None:
            stomp_email = self.opts["api_key_id"]
            stomp_password = self.opts["api_key_secret"]
        else:
            stomp_email = self.opts["email"]
            stomp_password = self.opts["password"]

        # Set up a STOMP connection to the Resilient action services
        stomp_timeout = int(self.opts.get("stomp_timeout")) # default from app.py:DEFAULT_STOMP_TIMEOUT
        # build out all the extra parameters for the stomp connections
        stomp_params = self.opts['resilient'].get('stomp_params')
        if not self.stomp_component:
            self.stomp_component = StompClient(stomp_host, self.opts["stomp_port"],
                                               username=stomp_email,
                                               password=stomp_password,
                                               heartbeats=(STOMP_CLIENT_HEARTBEAT,
                                                           STOMP_SERVER_HEARTBEAT),
                                               connected_timeout=stomp_timeout,
                                               connect_timeout=stomp_timeout,
                                               ssl_context=context,
                                               ca_certs=ca_certs,  # For old ssl version
                                               stomp_params=stomp_params,
                                               stomp_max_connection_errors=self.opts.get("stomp_max_connection_errors", constants.STOMP_MAX_CONNECTION_ERRORS),
                                               **self._proxy_args)
            self.stomp_component.register(self)
        else:
            # Component exists, just update it
            self.stomp_component.init(stomp_host, self.opts["stomp_port"],
                                      username=stomp_email,
                                      password=stomp_password,
                                      heartbeats=(STOMP_CLIENT_HEARTBEAT,
                                                  STOMP_SERVER_HEARTBEAT),
                                      connected_timeout=stomp_timeout,
                                      connect_timeout=stomp_timeout,
                                      ssl_context=context,
                                      ca_certs=ca_certs,  # For old ssl version
                                      stomp_params=stomp_params,
                                      stomp_max_connection_errors=self.opts.get("stomp_max_connection_errors", constants.STOMP_MAX_CONNECTION_ERRORS),
                                      **self._proxy_args)

        # Other special options
        self.ignore_message_failure = self.opts["resilient"].get("ignore_message_failure") == "1"
        self.fire(Event.create("reconnect", subscribe=False))

    def _setup_message_logging(self):
        """ Store action message logging option """
        self.logging_directory = self.opts["log_http_responses"] or None
        if self.logging_directory:
            try:
                directory = os.path.expanduser(self.logging_directory)
                directory = os.path.expandvars(directory)
                assert(os.path.exists(directory))
                self.logging_directory = directory
            except Exception as e:
                self.logging_directory = None
                raise Exception("Response Logging Directory %s does not exist!",
                                self.opts["log_http_responses"])

    @handler("registered")
    def registered(self, event, component, parent):
        """A component has registered.  Subscribe to its message queue(s)."""
        if self is component:
            self.reconnect_stomp = True
            self._setup_message_logging()
            self._setup_stomp()
            return
        # The set of channels for this component is:
        # - the component's channel(s) declared at class level, and
        # - any channel(s) declared at individual handlers
        channels = set(event.channels)
        for handler in component.handlers():
            if handler.channel:
                channels.update(handler.channel.split(","))
        for channel in channels:
            if str(channel).startswith("actions."):
                # Action module handler, channel "actions.xx" subscribes to "xx"
                queue_name = channel.split(".", 1)[1]
                LOG.info("'%s.%s' actions registered to '%s'",
                         type(component).__module__, type(component).__name__, queue_name)
            elif str(channel).startswith("functions.") and component._functions:
                if self.test_mode:
                    continue
                # Function handler, channel "functions.xx" subscribes to the message dest associated with function "xx"
                func_name = channel.split(".", 1)[1]
                if func_name not in component._functions:
                    # Unknown function.  Log it and continue.
                    LOG.warning("'%s.%s' function '%s' is not defined!",
                                type(component).__module__, type(component).__name__, func_name)
                    continue
                queue_name = component._functions[func_name]["destination_handle"]
                LOG.info("'%s.%s' function '%s' registered to '%s'",
                         type(component).__module__, type(component).__name__, func_name, queue_name)

            elif str(channel).startswith(constants.INBOUND_MSG_DEST_PREFIX):
                # If name for inbound q in app.config file, use that
                try:
                    app_config_q_name = component.app_configs.get(constants.INBOUND_MSG_APP_CONFIG_Q_NAME)
                except AttributeError as e:
                    raise IntegrationError(u"'{0}' does not have app_configs defined\n{1}".format(type(component).__name__, str(e)))
                if app_config_q_name:
                    handler_name = app_config_q_name
                else:
                    handler_name = channel.split(".", 1)[1]
                queue_name = u"{0}.{1}.{2}".format(constants.INBOUND_MSG_DEST_PREFIX, self.org_id, handler_name)
                LOG.info("'%s.%s' inbound handler '%s' registered to '%s'", type(component).__module__, type(component).__name__, handler_name, queue_name)

            else:
                continue

            if queue_name in self.listeners:
                comps = set([component])
                comps.update(self.listeners[queue_name])
                self.listeners[queue_name] = comps
            else:
                self.listeners[queue_name] = set([component])
                # Defer subscribing until all components are loaded

    @handler("load_all_success", "subscribe_to_all")
    def subscribe_to_queues(self):
        """ Subscribe to all message queues """
        if not self.stomp_component:
            return
        if not self.stomp_component.connected:
            yield self.wait("Connected", timeout=self.stomp_timeout)
            if not self.stomp_component.connected:
                LOG.error("Can't subscribe to queues with STOMP disconnected, trying reconnect")
                self.fire(Event.create("reconnect"))

        if self.stomp_component.connected:
            LOG.info("resilient-circuits has started successfully and is now running...")

        for queue_name in self.listeners:
            self._subscribe(queue_name)

    @handler("prepare_unregister")
    def prepare_unregister(self, event, component):
        """A component is unregistering.  Unsubscribe its message queue(s)."""
        LOG.debug("component %s has unregistered", component)
        if self is component:
            LOG.info("disconnecting Actions component from stomp queue")
            self.disconnect()
            self.reconnect_stomp = False
            if self.stomp_component:
                # TODO: Confirm the stomp component gets garbage collected automatically
                self.stomp_component.unregister()
                self.stomp_component = None

        for channel in event.channels:
            if not str(channel).startswith("actions."):
                continue
            LOG.info("Component %s unregistered from %s",
                     str(component), channel)
            queue_name = channel.partition(".")[2]
            if queue_name not in self.listeners:
                LOG.error("Channel %s was not subscribed", queue_name)
                continue
            comps = self.listeners[queue_name]
            if component not in comps:
                LOG.error("Component %s was not subscribed", component)
                continue
            comps.remove(component)
            if not comps:
                # All components have unsubscribed this destination; stop listening
                self._unsubscribe(queue_name)
            self.listeners[queue_name] = comps
            LOG.debug("Listeners: %s", self.listeners)

    def _subscribe(self, queue_name):
        """Actually subscribe the STOMP queue.  Note: this use client-ack, not auto-ack"""
        if self.resilient_mock:
            return

        if queue_name.startswith(constants.INBOUND_MSG_DEST_PREFIX):
            self.fire(Subscribe(queue_name, additional_headers=self.subscribe_headers))

        elif self.stomp_component and self.stomp_component.connected and self.listeners[queue_name]:
            if queue_name in self.stomp_component.subscribed:
                LOG.info("Ignoring request to subscribe to %s.  Already subscribed", queue_name)
            LOG.info("Subscribe to message destination '%s'", queue_name)

            if helpers.is_this_a_selftest(self):
                SELFTEST_SUBSCRIPTIONS.append(queue_name)

            destination = "actions.{0}.{1}".format(self.org_id, queue_name)
            self.fire(Subscribe(destination, additional_headers=self.subscribe_headers))
        else:
            LOG.error("Invalid request to subscribe to %s in state Connected? [%s] with %d listeners",
                      queue_name,
                      self.stomp_component.connected if self.stomp_component else "NO COMPONENT",
                      len(self.listeners[queue_name]))

    def _unsubscribe(self, queue_name):
        """Unsubscribe the STOMP queue"""
        if self.stomp_component and self.stomp_component.connected and self.listeners[queue_name]:
            LOG.info("Unsubscribe from message destination '%s'",
                     queue_name)
            destination = "actions.{0}.{1}".format(self.org_id, queue_name)
            self.fire(Unsubscribe(destination))

    def disconnect(self, reconnect=False, flush=True):
        """disconnect stomp connection"""
        if self.stomp_component:
            self.fire(Disconnect(reconnect=reconnect, flush=flush))

    @handler("reconnect")
    def _reconnect(self, subscribe=True):
        """Try (re)connect to the STOMP server"""
        if self.resilient_mock:
            return
        if self.stomp_component and self.stomp_component.connected and self.stomp_component.socket_connected:
            LOG.debug("STOMP reconnect requested when already connected")
        elif self.opts["resilient"].get("stomp") == "0":
            LOG.info("STOMP connection is not enabled")
        else:
            if self.stomp_component.socket_connected:
                LOG.warning("Disconnecting socket before Connect attempt")
                disconnect_event = Disconnect(reconnect=False, flush=False)
                self.fire(disconnect_event)
                yield self.wait(disconnect_event)

            LOG.info("STOMP attempting to connect")
            self.fire(Connect(subscribe=subscribe))

    @handler("Connect_success")
    def connected_succesfully(self, event, *args, **kwargs):
        """ Connected to stomp, subscribe if required """
        LOG.debug("Connected successfully. Resubscribe? %s", event.parent.subscribe)
        # This is used to automatically re-subscribe to queues on a reconnect
        if event.parent.subscribe:
            subscribe_event = Event.create("subscribe_to_all")
            self.fire(subscribe_event)
            yield self.wait(subscribe_event)

    @handler("Disconnected")
    @handler("ConnectionFailed")
    def retry_connection(self, event, *args, **kwargs):
        # Try again later
        reloading = getattr(self.parent, "reloading", False)
        if event.reconnect and not reloading:
            Timer(60, Event.create("reconnect")).register(self)

    @handler("exception")
    def exception(self, etype, value, traceback, handler=None, fevent=None):
        """Report an exception thrown during handling of an action event"""
        try:
            log_message, message = u"", u""
            if etype and issubclass(etype, BaseFunctionError):
                try:
                    message += str(value)
                except UnicodeDecodeError:
                    message += unicode(value)
                log_message = message
            else:
                if etype:
                    message = u"ERROR:\n{0}\n{1}".format(message, value)
                else:
                    message = u"Processing failed"
                if traceback and isinstance(traceback, list):

                    str_traceback = "".join(traceback)

                    if logging.getLogger().getEffectiveLevel() == logging.DEBUG:
                        message = u"{0}\n{1}".format(message, str_traceback)

                    log_message = u"{0}\n{1}".format(message, str_traceback)

            LOG.error(u"%s (%s): %s", repr(fevent), repr(etype), log_message)

            # Try find the underlying Action or Function message
            if fevent and fevent.args and not isinstance(fevent, ActionMessageBase):
                for arg in fevent.args:
                    if isinstance(arg, ActionMessageBase):
                        fevent = arg
                        break

            if isinstance(fevent, InboundMessage):
                headers = fevent.hdr()
                message_id = headers.get("message-id", None)
                if not fevent.test:
                    LOG.debug("Exception raised.\nAcknowledging InboundMessage: %s for queue: %s", message_id, headers.get("subscription", "Unknown"))
                    self.fire(Ack(fevent.frame))

            elif fevent and isinstance(fevent, ActionMessageBase):
                # For a ActionMessageBase type
                # 1. Ack message
                # 2. Send a message of type 1 containing the exception text
                fevent.stop()  # Stop further event processing
                status = 1
                headers = fevent.hdr()
                # Ack the message
                message_id = headers.get('message-id')
                if not fevent.test and self.stomp_component:
                    self.fire(Ack(fevent.frame, message_id=message_id))
                    LOG.debug("Ack %s", message_id)
                # Reply with error status
                reply_to = headers['reply-to']
                correlation_id = headers['correlation-id']
                reply_message = json.dumps({"message_type": status,
                                            "message": message,
                                            "complete": True})
                if not fevent.test and self.stomp_component:
                    self.fire(Send(headers={'correlation-id': correlation_id},
                                   body=reply_message,
                                   destination=reply_to,
                                   message_id=message_id))
                else:
                    # Test action, nothing to Ack
                    self.fire(Event.create("test_response",
                                           fevent.test_msg_id, reply_message))
                    LOG.debug("Test Action: No ack done.")
        except Exception as err:
            LOG.error("Exception handler threw exception! Response to action module may not have sent.")
            LOG.error(traceback)

    @handler("Ack_failure")
    def _on_ack_failure(self, event, err, *args, **kwargs):
        """STOMP Ack failed to send, add to delivery failures"""
        message_id = event.parent.message_id
        failure = self._stomp_ack_delivery_failures.get(message_id)
        if failure:
            if self.max_retry_count != 0 and failure["retry_count"] > self.max_retry_count:
                LOG.error("Giving up after %d attempts on delivery of STOMP ACK for message %s",
                          failure["retry_count"], message_id)
                self._stomp_ack_delivery_failures.pop(message_id)
        else:
            failure = {"retry_count": 1,
                       "from_prev_conn": False,
                       "message_frame": event.parent.frame}
            self._stomp_ack_delivery_failures[message_id] = failure

        LOG.warning("Failed %d times to deliver stomp ack for message %s", failure["retry_count"], message_id)

    @handler("Ack_success")
    def _on_ack_success(self, event, *args, **kwargs):
        if event.parent.message_id in self._stomp_ack_delivery_failures:
            LOG.info("Retry for sending STOMP ACK for message id %s successful.", event.parent.message_id)
            self._stomp_ack_delivery_failures.pop(event.parent.message_id)

    @handler("Send_success")
    def _on_send_success(self, event, *args, **kwargs):
        if event.parent.message_id in self._resilient_ack_delivery_failures:
            LOG.info("Retry for sending Resilient ACK for message id %s successful.", event.parent.message_id)
            self._resilient_ack_delivery_failures.pop(event.parent.message_id)

    @handler("Send_failure")
    def _on_send_failure(self, event, err, *args, **kwargs):
        """Resilient Ack failed to send, add to delivery failures"""
        message_id = event.parent.message_id

        failure = self._resilient_ack_delivery_failures.get(message_id)
        if failure:
            if self.max_retry_count != 0 and failure["retry_count"] > self.max_retry_count:
                LOG.error("Giving up after %d attempts on delivery of Resilient ACK for message %s",
                          failure["retry_count"], message_id)
                self._resilient_ack_delivery_failures.pop(message_id)
        else:
            failure = {"retry_count": 1,
                       "from_prev_conn": False,
                       "result":  {"headers": event.parent.headers,
                                   "body": event.parent.body,
                                   "destination": event.parent.destination}}
            self._resilient_ack_delivery_failures[message_id] = failure
        LOG.warning("Failed %d times to deliver Resilient ack for message %s", failure["retry_count"], message_id)

    @handler("retry_failed_deliveries")
    def _retry_send_failures(self, event):
        """retry all messages in the delivery_failures dict that are from the current session"""
        if not self.stomp_component:
            # Retries not applicable, probably using a mocked appliance
            return
        if not self.stomp_component.connected:
            LOG.info("Skipping retry of any failed messages because STOMP connection is down")
            return

        to_retry = {key: value for key, value in self._stomp_ack_delivery_failures.items()
                    if not value["from_prev_conn"]}
        for msgid, failure_info in to_retry.items():
            self._stomp_ack_delivery_failures[msgid]["retry_count"] = \
                self._stomp_ack_delivery_failures[msgid]["retry_count"] + 1
            LOG.info("Retrying failed STOMP ACK for message %s", msgid)
            self.fire(Ack(failure_info["message_frame"]))

        to_retry = {key: value for key, value in self._resilient_ack_delivery_failures.items()
                    if not value["from_prev_conn"]}
        for msgid, failure_info in to_retry.items():
            LOG.info("Retrying failed Resilient ACK for message %s", msgid)
            self._resilient_ack_delivery_failures[msgid]["retry_count"] = \
                self._resilient_ack_delivery_failures[msgid]["retry_count"] + 1
            self.fire(Send(headers=failure_info["result"]["headers"],
                           body=failure_info["result"]["body"],
                           destination=failure_info["result"]["destination"],
                           message_id=msgid))

    @handler("signal")
    def _on_signal(self, signo, stack):
        """We implement a default-event handler, which means we don't get default signal handling - add it back
           (see FallBackSignalHandler in circuits/core/helpers.py)
        """
        if signo in [SIGINT, SIGTERM]:
            raise SystemExit(0)

    @handler("SelftestTerminateEvent")
    def _selftest_terminate(self):
        """
        Exits resilient-circuits if a SelftestTerminateEvent
        is fired
        """
        LOG.info("SelftestTerminateEvent, exiting resilient-circuits")
        raise SystemExit(0)

    @handler("reload", priority=999)
    def reload(self, event, opts):
        """New config, check to see if num_workers has been changed, update pool if it has and is valid"""
        reloaded_num_workers = opts.get("num_workers", -1)
        if constants.MIN_NUM_WORKERS <= reloaded_num_workers <= constants.MAX_NUM_WORKERS:
            if reloaded_num_workers != self._num_workers:
                LOG.debug("The num_workers app.config setting has been changed from %s to %s.",
                          self._num_workers, reloaded_num_workers)
                # Unregister the existing worker thread-pool that will run functions.
                self._functionworker.unregister()
                # Re-create and register the new worker thread-pool that will run functions with new num_workers setting.
                self._functionworker = FunctionWorker(process=False, channel="functionworker",
                                                      workers=reloaded_num_workers)
                self._functionworker.register(self.root)
                # Update the instance attribute.
                self._num_workers = reloaded_num_workers
        else:
            LOG.error("The num_workers app.config setting has been changed to an invalid value %s",
                      reloaded_num_workers)

        """New config, reconnect to stomp if required"""
        event.success = False
        super(Actions, self).reload(event, opts)
        self._configure_opts(opts)
        if self.stomp_component:
            self.fire(Disconnect(flush=True, reconnect=False))
            yield self.wait("Disconnect_success")
            self._setup_stomp()
            yield self.wait("Connect_success")
            subscribe_event = Event.create("subscribe_to_all")
            self.fire(subscribe_event)
            yield self.wait(subscribe_event)
        event.success = True

    @handler("StatusMessageEvent", "FunctionErrorEvent")
    def _status(self, event, *args, **kwargs):
        """Report an interim status"""
        if not isinstance(event.parent, ActionMessageBase):
            return
        fevent = event.parent
        # Reply with interim status
        message = event.text
        if not message:
            message = "(No status)"
        if isinstance(event, FunctionErrorEvent):
            status = 1
            complete = True
            fevent.stop()  # Stop further event processing
        else:
            status = 0
            complete = False
        headers = fevent.hdr()
        message_id = headers.get('message-id', None)
        reply_to = headers['reply-to']
        correlation_id = headers['correlation-id']
        reply_message = json.dumps({"message_type": status,
                                    "message": message,
                                    "complete": complete})
        if not fevent.test:
            self.fire(Send(headers={'correlation-id': correlation_id},
                           body=reply_message,
                           destination=reply_to,
                           message_id=message_id))
        else:
            # Test action, nothing to Ack
            # Send a message back to the test client
            self.fire(Event.create("test_response", fevent.test_msg_id, reply_message), '*')
            LOG.debug("Test Action: No ack done.")

    @handler()
    def _on_event(self, event, *args, **kwargs):
        """Report the successful handling of an action event"""
        if isinstance(event.parent, ActionMessageBase) and event.name.endswith("_success"):
            fevent = event.parent
            headers = fevent.hdr()
            message_id = headers.get("message-id", None)

            if fevent.deferred:
                LOG.debug("Not acking deferred message %s", str(fevent))

            elif isinstance(fevent, InboundMessage):
                if not fevent.test:
                    LOG.debug("Acknowledging InboundMessage: %s for queue: %s", message_id, headers.get("subscription", "Unknown"))
                    self.fire(Ack(fevent.frame))
            else:
                value = event.parent.value.getValue()
                LOG.debug("success! %s, %s", value, fevent)
                fevent.stop()  # Stop further event processing

                # At completion, the value might be:
                # - None (if there was no handler or a handler just returned None)
                # - or a single thing, or a list; where each thing could be
                #   - a StatusMessage
                #   - a FunctionResult
                #   - a plain Python object: string, number, dictionary, etc
                # Custom Actions don't have any "result", so all the values are
                # concatenated into a status message.
                # Functions have a result as well as a status message.
                if not isinstance(value, list):
                    value = [value]
                function_result = None
                status_messages = []

                if isinstance(event.parent, ActionMessage):
                    # All the values become the status message.
                    for val in value:
                        if isinstance(val, StatusMessage):
                            status_messages.append(val)
                        elif val is not None and not isinstance(val, FunctionResult):
                            status_messages.append(StatusMessage(val))
                    # If the function didn't return a status message, let's make one.
                    if not status_messages:
                        status_messages.append(StatusMessage(u"No handler returned a result for this action"))

                if isinstance(event.parent, FunctionMessage):
                    # The first (and only the first!) FunctionResult is the result.
                    for val in value:
                        if isinstance(val, FunctionResult):
                            function_result = val
                            break
                    # If there was no FunctionResult, but there was a plain value, convert that to a FunctionResult
                    values = []
                    for val in value:
                        if function_result is None and \
                                not isinstance(val, FunctionResult) and \
                                not isinstance(val, StatusMessage):
                            function_result = FunctionResult(val)
                        else:
                            values.append(val)
                    # All the other values become the status message.
                    for val in values:
                        if isinstance(val, FunctionResult):
                            pass
                        elif isinstance(val, StatusMessage):
                            status_messages.append(val)
                        elif isinstance(val, string_types):
                            status_messages.append(StatusMessage(val))
                    # If the function didn't return a status message, let's make one.
                    if not status_messages:
                        status_messages.append(StatusMessage(u"Completed"))

                messages = []
                for status in status_messages:
                    try:
                        messages.append(str(status))
                    except:
                        messages.append(repr(status))
                message = "\n".join(messages)
                LOG.debug(u"Message: %s", message)
                status = 0
                headers = fevent.hdr()
                # Ack the message
                message_id = headers.get('message-id', None)
                if not fevent.test:
                    LOG.debug("Ack %s", message_id)
                    self.fire(Ack(fevent.frame, message_id=message_id))
                # Reply with success status
                reply_to = headers['reply-to']
                correlation_id = headers['correlation-id']
                # reply_message is an ActionAcknowledgementDTO
                reply_dto = {"message_type": status,
                             "message": message,
                             "complete": True}
                if function_result:
                    LOG.debug("[%s] Result: %s", function_result.name, function_result.value)
                    reply_dto["results"] = function_result.value
                reply_message = json.dumps(reply_dto, indent=2)
                if not fevent.test:
                    self.fire(Send(headers={'correlation-id': correlation_id},
                                   body=reply_message,
                                   destination=reply_to,
                                   message_id=message_id))
                else:
                    # Test action, nothing to Ack
                    if isinstance(event.parent, FunctionMessage):
                        # Send the value back to the test
                        result = Event.create("{}_result".format(event.parent.name), result=function_result)
                        result.parent = event.parent
                        self.fire(result, '*')
                    # Send a message back to the test client
                    self.fire(Event.create("test_response", fevent.test_msg_id, reply_message), '*')
                    LOG.debug("Test Action: No ack done.")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# (c) Copyright IBM Corp. 2010, 2021. All Rights Reserved.

"""Common constants used in tests"""

from resilient_circuits import helpers, constants

MOCK_ORG = "201"
MOCK_QUEUE_NAME = "mock_queue"
MOCK_SUBSCRIPTION = "{0}.{1}.{2}".format(constants.INBOUND_MSG_DEST_PREFIX, MOCK_ORG, MOCK_QUEUE_NAME)
MOCK_DESTINATION = "/queue/{0}".format(MOCK_SUBSCRIPTION)
MOCK_QUEUE = helpers.get_queue(MOCK_DESTINATION)

MOCK_PACKAGE_NAME = u"mock_function_package"

MOCK_FUNCTION_PREFIX = "function_mock"
MOCK_FN_NAME_ONE = u"{0}_{1}".format(MOCK_FUNCTION_PREFIX, "one")
MOCK_FN_NAME_EX = u"{0}_{1}".format(MOCK_FUNCTION_PREFIX, "raise_exception")

MOCK_INBOUND_Q_NAME = u"mock_inbound_q_name"
MOCK_INBOUND_Q_NAME_CREATE = u"{0}_{1}".format(MOCK_INBOUND_Q_NAME, "create")
MOCK_INBOUND_Q_NAME_EX = u"{0}_{1}".format(MOCK_INBOUND_Q_NAME, "raise_exception")

MOCK_APP_FUNCTION_PREFIX = "app_function_mock"
MOCK_APP_FN_NAME_ONE = u"{0}_{1}".format(MOCK_APP_FUNCTION_PREFIX, "one")
MOCK_APP_FN_NAME_CUSTOM_RESULT = u"{0}_{1}".format(MOCK_APP_FUNCTION_PREFIX, "custom_result")
MOCK_APP_FN_NAME_EX = u"{0}_{1}".format(MOCK_APP_FUNCTION_PREFIX, "raise_exception")

RESILIENT_MOCK = u"pytest_resilient_circuits.BasicResilientMock"

CONFIG_DATA = """[{0}]
url = https://www.example.com""".format(MOCK_PACKAGE_NAME)

MOCK_REQUIRED_APP_CONFIGS = [
    {"name": "url", "placeholder": "https://www.example.com"}
]

MOCK_APP_CONFIGS = {
    "url": "https://www.mockexample.com"
}

MOCK_OPTS = {
    "host": "192.168.0.1",
    "org": "Test Organization",
    "email": "admin@example.com",
    "password": "123",
    "resilient_mock": RESILIENT_MOCK,
    MOCK_PACKAGE_NAME: MOCK_APP_CONFIGS
}



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# (c) Copyright IBM Corp. 2010, 2018. All Rights Reserved.

"""Action Module server - restartable version of App"""

from __future__ import print_function

import logging
import os

import filelock
from circuits import Event, Timer
from resilient_circuits.app import (App, AppArgumentParser,
                                    RotatingFileHandler, constants, get_lock,
                                    helpers)
from watchdog.events import PatternMatchingEventHandler
from watchdog.observers import Observer

import resilient

application = None
LOG = logging.getLogger(__name__)


class reload(Event):
    """Notify components of updates to config"""
    complete = True

    def __init__(self, opts):
        super(reload, self).__init__(opts=opts)
        self.opts = opts


class ConfigFileUpdateHandler(PatternMatchingEventHandler):
    """ Restarts application when config file is modified """

    def __init__(self, app):
        super(ConfigFileUpdateHandler, self).__init__()
        self.app = app
        self.max_reload_time = 30

    @classmethod
    def set_patterns(cls, config_file):
        cls.patterns = ["*" + os.path.basename(config_file), ]

    def on_modified(self, event):
        """ For 'FileModifiedEvent' events, initiate reload of data from config file and restart components.  """

        self.reload_config()

    def on_created(self, event):
        """ For 'FileCreatedEvent' events, initiate reload of data from config file and restart components.

        Triggered by certain editors such as 'vi', which will recreate rather than modify the config file,
        when updating.

        """

        self.reload_config()

    def reload_config(self):
        """ Reload data from config file and restart components. """

        if self.app.reloading:
            LOG.warning("Configuration file change ignored because reload already in progress")
            return

        LOG.info("Configuration file has changed! Notify components to reload")
        self.app.reloading = True
        opts = helpers.get_configs(path_config_file=self.app.config_file, ALLOW_UNRECOGNIZED=self.app.ALLOW_UNRECOGNIZED)
        # See if we need to reset root loglevel on reload
        self.reset_loglevel(opts)
        # check if we need to reset rotating file handler opts
        self.reset_filehandler(opts)
        reload_event = reload(opts=opts)
        self.app.reload_timer = Timer(self.max_reload_time, Event.create("reload_timeout"))
        self.app.fire(reload_event)
        self.app.reload_timer.register(self.app)

    def reset_loglevel(self, opts):
        """ Reset the loglevel if it has changed in the config file. """

        # Get loglevel setting from app.config.
        config_loglevel = opts.get("loglevel", None)

        # If loglevel setting is configured.
        if config_loglevel:
            # Check if it's a valid setting
            if hasattr(logging, config_loglevel.upper()):
                # Get root logger instance
                root_logger = logging.getLogger()
                # If level is different set loglevel to new value.
                if logging.getLevelName(root_logger.level) != config_loglevel.upper():
                    root_logger.setLevel(config_loglevel.upper())
            else:
                LOG.error("Invalid app.config setting for loglevel %s", config_loglevel)

    def reset_filehandler(self, opts):
        """
        Reset ``log_max_bytes`` and ``log_backup_count`` if changed on reload
        """

        log_max_bytes = opts.get(constants.APP_CONFIG_LOG_MAX_BYTES)
        log_backup_count = opts.get(constants.APP_CONFIG_LOG_BACKUP_COUNT)

        # loop through handlers and find the RotatingFileHandler
        # (it will be there -- just need to find it in the list)
        handlers = getattr(logging.getLogger(), "handlers", [])
        for handler in handlers:
            if isinstance(handler, RotatingFileHandler):

                # update values as appropriate
                if log_backup_count is not None and handler.backupCount != log_backup_count:
                    LOG.debug("Reloaded 'log_backup_count' to '%s'",log_backup_count)
                    handler.backupCount = log_backup_count
                if log_max_bytes is not None and handler.maxBytes != log_max_bytes:
                    LOG.debug("Reloaded 'log_max_bytes' to '%s'", log_max_bytes)
                    handler.maxBytes = log_max_bytes
                break # break the loop as we only needed the RotatingFileHandler

# Main component for our application
class AppRestartable(App):
    """Our main app component, which sets up the Resilient services and other components"""

    def __init__(self, *args, **kwargs):
        super(AppRestartable, self).__init__(*args, **kwargs)
        self.reloading = False
        self.reload_timer = None
        self.observer = None

    def do_initialize_watchdog(self):
        """Initialize the configuration file watchdog"""
        # Monitor the configuration file, using a Watchdog observer daemon.
        LOG.info("Monitoring config file for changes.")
        ConfigFileUpdateHandler.set_patterns(self.config_file)
        event_handler = ConfigFileUpdateHandler(self)
        self.observer = Observer()
        config_dir = os.path.dirname(self.config_file)
        if not config_dir:
            config_dir = os.getcwd()
        self.observer.schedule(event_handler, path=config_dir, recursive=False)
        self.observer.daemon = True
        self.observer.start()

    def started(self, component):
        LOG.info("App Started %s", str(component))
        self.do_initialize_watchdog()

    def stopped(self, component):
        """Stopped Event Handler"""
        LOG.info("App Stopped")
        self._stop_observer()

    def reload_complete(self, event, *args, **kwargs):
        """ All components done handling reload event """
        if event.parent.success:
            LOG.info("Reload completed successfully!")
        else:
            LOG.error("Reloading failed to complete successfully!")
        if self.reload_timer:
            self.reload_timer.unregister()
            self.reload_timer = None
        self.reloading = False

    def reload_timeout(self, event):
        """Reload timed out, assume it failed"""
        LOG.error("Reload Timed Out, Assuming Failure!")
        self.reloading = False

    def _stop_observer(self):
        """ stop monitoring config file for changes """
        if self.observer:
            LOG.info("Stopping config file monitoring")
            self.observer.unschedule_all()
            self.observer.stop()
            self.observer = None


def run(*args, **kwargs):
    """Main app"""

    # define lock
    # this prevents multiple, identical circuits from running at the same time
    lock = get_lock()

    # The main app component initializes the Resilient services
    global application
    try:
        # attempt to lock file, wait 1 second for lock
        with lock.acquire(timeout=1):
            assert lock.is_locked
            application = AppRestartable(*args, **kwargs)
            application.run()

    except filelock.Timeout:
        # file is probably already locked
        errmsg = ("Failed to acquire lock on {0} - you may have "
                  "another instance of Resilient Circuits running")
        print(errmsg.format(os.path.abspath(lock.lock_file)))
    except ValueError:
        LOG.exception("ValueError Raised. Application not running.")
    except OSError as exc:
        # Some other problem accessing the lockfile
        print("Unable to lock {0}: {1}".format(os.path.abspath(lock.lock_file), exc))
    # finally:
    #    LOG.info("App finished.")


if __name__ == "__main__":
    run()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# (c) Copyright IBM Corp. 2010, 2021. All Rights Reserved.

import logging
import os
import time
from collections import defaultdict
from threading import Thread

import pkg_resources
from requests.exceptions import ConnectionError, SSLError
from resilient import BasicHTTPException, SimpleClient, SimpleHTTPException
from resilient import constants as res_constants
from resilient import get_and_parse_proxy_env_var, is_env_proxies_set
from resilient_circuits import app, constants, helpers
from resilient_circuits.actions_component import (SELFTEST_ERRORS,
                                                  SELFTEST_SUBSCRIPTIONS)
from resilient_circuits.stomp_events import SelftestTerminateEvent

# Get the same logger object that is used for resilient_circuits commands
LOG = logging.getLogger(constants.CMDS_LOGGER_NAME)

ERROR_EXIT_CODES_MAP = {
    1: constants.EXIT_SELFTEST_ERROR,
    2: constants.EXIT_SELFTEST_UNIMPLEMENTED,
    20: constants.EXIT_REST_CONNECTION_ERROR,
    401: constants.EXIT_REST_UNAUTHORIZED,
    22: constants.EXIT_REST_OS_ERROR,
    23: constants.EXIT_REST_SSL_ERROR,
    24: constants.EXIT_REST_ORG_ERROR,
    25: constants.EXIT_REST_INVALID_PW,
    30: constants.EXIT_STOMP_ERROR,
    31: constants.EXIT_STOMP_UNAUTHORIZED_CONN,
    32: constants.EXIT_STOMP_UNAUTHORIZED_Q,
    33: constants.EXIT_STOMP_Q_TIMEOUT
}


def error_connecting_to_soar(host, reason="Unknown", status_code=20):
    """
    Logs the host, reason and error code why it cannot connect to SOAR
    and exits with the error code defined in ERROR_EXIT_CODES_MAP
    """

    # Remove resilient-circuits logging handler
    LOG.parent.handlers = []

    LOG.info("\nERROR: could not connect to SOAR at '{0}'.\nReason: {1}\nError Code: {2}".format(host, reason, ERROR_EXIT_CODES_MAP.get(status_code, 1)))
    exit(ERROR_EXIT_CODES_MAP.get(status_code, 1))


def check_soar_rest_connection(cmd_line_args, app_configs):
    """
    Check if we can  successfully get a resilient_client
    therefore that will tell us if we have configured the app.config
    file correctly in order to establish a REST connection and authenticate
    with SOAR

    :param cmd_line_args: an argparse.Namespace object containing all command line params
    :type cmd_line_args: argparse.Namespace
    :param app_configs: a dict of all the configurations in the app.config file
    :type app_configs: dict
    :excepts BasicHTTPException: if we cannot authenticate. Exits with 21
    :excepts SSLError: if the cafile that is supplied is invalid. Exits with 23
    :excepts Exception: generic error. Also raises if the user is not a member of the current org. Exits with 20
    :return: Nothing

    """
    LOG.info("{0}Testing REST connection to SOAR{0}".format(constants.LOG_DIVIDER))

    host = app_configs.get("host", constants.DEFAULT_NONE_STR)
    user = app_configs.get("api_key_id", app_configs.get("email", constants.DEFAULT_NONE_STR))
    cafile = app_configs.get("cafile") if app_configs.get("cafile") else ""

    if not os.path.isfile(cafile):
        LOG.warning("- WARNING: No certificate file specified. Only allows the connections that are trusted by operating system.")

    LOG.info("- Checking if we can authenticate a REST connection with '{0}' to '{1}'".format(user, host))

    if is_env_proxies_set():
        proxy_details = get_and_parse_proxy_env_var(res_constants.ENV_HTTPS_PROXY)

        if not proxy_details:
            proxy_details = get_and_parse_proxy_env_var(res_constants.ENV_HTTP_PROXY)

        LOG.info("- Using a '{0}' Proxy with Host '{1}' and Port '{2}'".format(proxy_details.get("scheme"), proxy_details.get("hostname"), proxy_details.get("port")))

    try:
        res_client = helpers.get_resilient_client(ALLOW_UNRECOGNIZED=True)

    except BasicHTTPException as e:
        # Connection unauthorized
        error_connecting_to_soar(host, e.response.reason, e.response.status_code)

    except SimpleHTTPException as e:
        # Incorrect User/Password
        if hasattr(e, "args") and isinstance(e.args, tuple) and constants.ERROR_INVALID_USR in e.args[0]:
            error_connecting_to_soar(host, e, 25)

        error_connecting_to_soar(host, u"Unknown REST Error: {0}".format(e), 20)

    except SSLError as e:
        # SSL Error (Certificate Error)
        error_connecting_to_soar(host, e, 23)

    except ConnectionError as e:
        # Generic connection error
        error_connecting_to_soar(host, e, 20)

    except OSError as e:
        if hasattr(e, "args") and isinstance(e.args, tuple) and constants.ERROR_CA_FILE_NOT_FOUND in e.args[0]:
            # CA file could not be found/read
            error_connecting_to_soar(host, e, 22)

        error_connecting_to_soar(host, u"Unknown REST Error: {0}".format(e), 20)

    except Exception as e:
        # Generic connection error (normally related to the user's org membership)
        if hasattr(e, "args") and isinstance(e.args, tuple):

            # User not member of organization error
            if constants.ERROR_USR_NOT_MEMBER_ORG in e.args[0]:
                error_connecting_to_soar(host, e, 24)

            error_connecting_to_soar(host, e, 20)

        error_connecting_to_soar(host, u"Unknown REST Error: {0}".format(e), 20)

    if not isinstance(res_client, SimpleClient):
        error_connecting_to_soar(host, "Unknown REST Error", 20)

    LOG.info("{0}Successfully connected via REST!{0}".format(constants.LOG_DIVIDER))


def check_soar_stomp_connection(cmd_line_args, app_configs):
    """
    Check if we can successfully start an instance of resilient_circuits
    and therefore that will tell us if we have configured the app.config
    file correctly in order to establish a STOMP connection and authenticate
    with SOAR

    :param cmd_line_args: an argparse.Namespace object containing all command line params
    :type cmd_line_args: argparse.Namespace
    :param app_configs: a dict of all the configurations in the app.config file
    :type app_configs: dict
    :excepts BasicHTTPException: if we cannot authenticate. Exits with 32
    :excepts Exception: generic error. Exits with 30
    :return: Nothing

    """
    LOG.info("{0}Testing STOMP connection to SOAR{0}".format(constants.LOG_DIVIDER))

    host = app_configs.get("host", constants.DEFAULT_NONE_STR)
    user = app_configs.get("api_key_id", app_configs.get("email", constants.DEFAULT_NONE_STR))

    LOG.info("- Checking if we can authenticate a STOMP connection with '{0}' to '{1}'".format(user, host))

    try:
        LOG.info("{0}Instantiating instance of resilient-circuits and starting it...{0}".format(constants.LOG_DIVIDER))
        resilient_circuits_instance = app.App(ALLOW_UNRECOGNIZED=True, IS_SELFTEST=True)

        # Create thread that targets the `run()` method of the main resilient_circuits component
        t_running_resilient_circuits = Thread(target=resilient_circuits_instance.run, name="resilient_circuits")
        t_running_resilient_circuits.start()

        start_time = time.time()

        while len(SELFTEST_SUBSCRIPTIONS) == 0:
            LOG.parent.info("- Waiting for subscription to message destination. Sleeping for 2 seconds")
            time.sleep(2)

            if helpers.should_timeout(start_time, app_configs.get(constants.APP_CONFIG_SELFTEST_TIMEOUT, constants.DEFAULT_SELFTEST_TIMEOUT_VALUE)):
                resilient_circuits_instance.action_component.fire(SelftestTerminateEvent())
                error_connecting_to_soar(host, "Could not subscribe to any message destinations", 33)

        # Send event to Terminate resilient-circuits
        resilient_circuits_instance.action_component.fire(SelftestTerminateEvent())

        # Remove resilient-circuits logging handler
        time.sleep(2)
        LOG.parent.handlers = []

        if SELFTEST_ERRORS:
            # Not authorized to read from queue
            for e in SELFTEST_ERRORS:
                if b"is not authorized to read from queue" in e:
                    error_connecting_to_soar(host, "'{0}' is not authorized to read from the App's Message Destination".format(helpers.get_user(app_configs)), 32)

            error_connecting_to_soar(host, u"Unknown STOMP Error: {0}".format(e), 30)

    except BasicHTTPException as e:
        # Not authorized to instansiate STOMP connection
        error_connecting_to_soar(host, e.response.reason, 31)

    except Exception as e:
        # Generic connection error
        error_connecting_to_soar(host, u"Unknown STOMP Error: {0}".format(e), 30)

    LOG.info("{0}Successfully connected via STOMP!{0}".format(constants.LOG_DIVIDER))


def run_apps_selftest(cmd_line_args, app_configs):
    """
    loop through every selftest for every eligible package, call and store returned state,
    print out package and their selftest states
    """

    if hasattr(cmd_line_args, "print_env") and cmd_line_args.print_env:
        LOG.info(helpers.get_env_str(pkg_resources.working_set))

    components = defaultdict(list)

    # custom entry_point only for selftest functions
    selftest_entry_points = [ep for ep in pkg_resources.iter_entry_points('resilient.circuits.selftest')]
    for ep in selftest_entry_points:
        components[ep.dist].append(ep)

    if len(selftest_entry_points) == 0:
        LOG.info("No selftest entry points found.")
        return None

    # make a copy
    install_list = list(cmd_line_args.install_list) if cmd_line_args.install_list else []

    # Prepare a count of exceptions found with selftests.
    selftest_failure_count = 0
    selftest_unimplemented_count = 0
    selftest_not_found_count = 0
    selftest_unknown_count = 0

    for dist, component_list in components.items():
        if cmd_line_args.install_list is None or dist.project_name in install_list:
            # remove name from list
            if dist.project_name in install_list:
                install_list.remove(dist.project_name)

            LOG.info("{0}Running selftest for: '{1}'{0}".format(constants.LOG_DIVIDER, dist.project_name))
            # add an entry for the package
            LOG.info("\n%s: ", dist.project_name)
            for ep in component_list:
                # load the entry point
                try:
                    f_selftest = ep.load()
                except ModuleNotFoundError:
                    selftest_not_found_count += 1
                    continue

                try:
                    # f_selftest is the selftest function, we pass the selftest resilient options in case it wants to use it
                    start_time_milliseconds = int(round(time.time() * 1000))

                    status = f_selftest(app_configs)

                    end_time_milliseconds = int(round(time.time() * 1000))

                    delta_milliseconds = end_time_milliseconds - start_time_milliseconds
                    delta_seconds = delta_milliseconds / 1000

                    state = status.get("state")

                    if isinstance(state, str):
                        LOG.info("\t%s: %s\n\tselftest output:\n\t%s\n\tElapsed time: %f seconds", ep.name, state, status, delta_seconds)

                        if state.lower() == constants.SELFTEST_FAILURE_STATE:
                            selftest_failure_count += 1
                        elif state.lower() == constants.SELFTEST_UNIMPLEMENTED_STATE:
                            selftest_unimplemented_count += 1
                        elif state.lower() != constants.SELFTEST_SUCCESS_STATE:
                            selftest_unknown_count += 1

                    else:
                        LOG.info("\t%s:\n\tUnsupported dictionary returned:\n\t%s\n\tElapsed time: %f seconds", ep.name, status, delta_seconds)

                except Exception as e:
                    LOG.error("Error while calling %s. Exception: %s", ep.name, str(e))
                    selftest_failure_count += 1
                    continue

    # any missed packages?
    if install_list:
        LOG.warning("\n%s not found. Selftest not ran for %s. Check package name(s)\n", install_list, install_list)

    # Check if any failures were found and printed to the console
    if selftest_failure_count or selftest_unknown_count:
        LOG.info("\nERROR: running selftest for App.\nError Code: {0}".format(ERROR_EXIT_CODES_MAP.get(1, 1)))
        exit(ERROR_EXIT_CODES_MAP.get(1, 1))
    elif selftest_unimplemented_count:
        LOG.info("\nERROR: selftest is unimplemented for this App. Note: the App may still continue to work...\nError Code: {0}".format(ERROR_EXIT_CODES_MAP.get(2, 2)))
        exit(ERROR_EXIT_CODES_MAP.get(2, 2))
    elif selftest_not_found_count:
        LOG.info("\nERROR: selftest not found for this App. Note: the App may still continue to work...\nError Code: {0}".format(ERROR_EXIT_CODES_MAP.get(2, 2)))
        exit(ERROR_EXIT_CODES_MAP.get(2, 2))

    if not install_list:
        LOG.info("{0}Successfully ran App's selftest!{0}".format(constants.LOG_DIVIDER))


def execute_command(cmd_line_args):
    LOG.info("{0}Running selftest with IBM SOAR{0}".format(constants.LOG_DIVIDER))

    if hasattr(cmd_line_args, "print_env") and cmd_line_args.print_env:
        LOG.info("- Printing runtime environment")
        LOG.info(helpers.get_env_str(pkg_resources.working_set))

    LOG.info("- Getting app.configs")
    app_configs = helpers.get_configs(ALLOW_UNRECOGNIZED=True)

    check_soar_rest_connection(cmd_line_args, app_configs)
    check_soar_stomp_connection(cmd_line_args, app_configs)
    run_apps_selftest(cmd_line_args, app_configs)

    LOG.info("{0}selftest complete{0}".format(constants.LOG_DIVIDER))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# (c) Copyright IBM Corp. 2010, 2022. All Rights Reserved.

import os

import pkg_resources

PACKAGE_NAME = "resilient-circuits"

PASSWD_PATTERNS = ['token', 'pass', 'secret', 'pin', 'key', 'id']

INBOUND_MSG_DEST_PREFIX = "inbound_destinations"

APP_FUNCTION_PAYLOAD_VERSION = 2.0

MIN_NUM_WORKERS = 1
MAX_NUM_WORKERS = 500
MIN_LOG_BYTES = 100000
MIN_BACKUP_COUNT = 0
DEFAULT_SELFTEST_TIMEOUT_VALUE = 10

APP_LOG_DIR = os.environ.get("APP_LOG_DIR", "logs")
CMDS_LOGGER_NAME = "resilient_circuits_cmd_logger"
LOG_DIVIDER = "\n------------------------\n"

DEFAULT_NONE_STR = "Not found"
DEFAULT_UNKNOWN_STR = "Unknown"
ERROR_CA_FILE_NOT_FOUND = "Could not find a suitable TLS CA certificate bundle"
ERROR_USR_NOT_MEMBER_ORG = "The user is not a member of the specified organization"
ERROR_INVALID_USR = "Invalid user name or password"

# Selftest
SELFTEST_SUCCESS_STATE = "success"
SELFTEST_FAILURE_STATE = "failure"
SELFTEST_UNIMPLEMENTED_STATE = "unimplemented"

# app configs
INBOUND_MSG_APP_CONFIG_Q_NAME = "inbound_destination_api_name"
APP_CONFIG_TRAP_EXCEPTION = "trap_exception"
APP_CONFIG_SELFTEST_TIMEOUT = "selftest_timeout"
APP_CONFIG_HEARTBEAT_TIMEOUT_THRESHOLD = "heartbeat_timeout_threshold"
APP_CONFIG_LOG_MAX_BYTES = "log_max_bytes"
APP_CONFIG_LOG_BACKUP_COUNT = "log_backup_count"

# Headers
HEADER_CIRCUITS_VER_KEY = "Resilient-Circuits-Version"
HEADER_CIRCUITS_VER_VALUE = pkg_resources.get_distribution(PACKAGE_NAME).version

# Exit Codes
EXIT_SELFTEST_ERROR = 1             # Error running App's selftest
EXIT_SELFTEST_UNIMPLEMENTED = 2     # selftest was unimplemented
EXIT_REST_CONNECTION_ERROR = 20     # REST: Generic connection error
EXIT_REST_UNAUTHORIZED = 21         # REST: Connection unauthorized
EXIT_REST_OS_ERROR = 22             # REST: OSError (Could not find Certificate file)
EXIT_REST_SSL_ERROR = 23            # REST: SSL Error (Invalid Certificate Error)
EXIT_REST_ORG_ERROR = 24            # REST: Organization Membership Error
EXIT_REST_INVALID_PW = 25           # REST: Invalid Username or Password
EXIT_STOMP_ERROR = 30               # STOMP: Generic connection error
EXIT_STOMP_UNAUTHORIZED_CONN = 31   # STOMP: Not authorized to instansiate STOMP connection
EXIT_STOMP_UNAUTHORIZED_Q = 32      # STOMP: Not authorized to read from queue
EXIT_STOMP_Q_TIMEOUT = 33           # STOMP: Timed out trying to see if resilient-circuits is subscribed to a message destination
EXIT_STOMP_HEARTBEAT_TIMEOUT = 34   # STOMP: The delta of the current HeartbeatTimeout event and the first HeartbeatTimeout event is greater than the 'heartbeat_timeout_threshold'

# STOMP connection constants
STOMP_MAX_CONNECTION_ERRORS = 1     # default number of errors when heartbeat is lost


#!/usr/bin/env python
# -*- coding: utf-8 -*-
# (c) Copyright IBM Corp. 2010, 2021. All Rights Reserved.

import pytest
import copy
from resilient_circuits.component_loader import ComponentLoader
from resilient_circuits import constants, helpers
from tests import mock_constants, mock_paths


class TestComponentLoader:

    @pytest.mark.parametrize("ep_str", ["mock_component:MockInboundAppComponent"])
    @pytest.mark.parametrize("path_dist", [mock_paths.SHARED_MOCK_DATA_DIR])
    def test_discover_installed_components_with_inbound_app(self, fx_add_entry_point):
        found = False

        cmp_loader_opts = copy.deepcopy(mock_constants.MOCK_OPTS)
        cmp_loader_opts["componentsdir"] = mock_paths.SHARED_MOCK_DATA_DIR

        cmp_loader = ComponentLoader(cmp_loader_opts)
        installed_cmps = cmp_loader.discover_installed_components()

        assert len(installed_cmps) > 0

        for c in installed_cmps:
            if c.__module__ == "mock_component":
                found = True

        assert found is True

    @pytest.mark.parametrize("ep_str", ["mock_component:MockInboundAppComponent"])
    @pytest.mark.parametrize("path_dist", [mock_paths.SHARED_MOCK_DATA_DIR])
    def test_installed_components_with_custom_inbound_q(self, fx_add_entry_point):
        found = False
        custom_q_name = "mock_inbound_custom_q"

        cmp_loader_opts = copy.deepcopy(mock_constants.MOCK_OPTS)
        cmp_loader_opts["componentsdir"] = mock_paths.SHARED_MOCK_DATA_DIR

        cmp_loader_opts["mock_component"] = {}
        cmp_loader_opts["mock_component"][constants.INBOUND_MSG_APP_CONFIG_Q_NAME] = custom_q_name

        cmp_loader = ComponentLoader(cmp_loader_opts)
        installed_cmps = cmp_loader.discover_installed_components()

        assert len(installed_cmps) > 0

        for c in installed_cmps:
            if c.__module__ == "mock_component":

                inbound_handlers = helpers.get_handlers(c, handler_type="inbound_handler")
                assert len(inbound_handlers) > 0

                for ih in inbound_handlers:
                    if ih[0] == "_inbound_app_mock_one":
                        found = True
                        assert ih[1].channel == "{0}.{1}".format(constants.INBOUND_MSG_DEST_PREFIX, custom_q_name)

        assert found is True





def selftest(opts):
    """
    Placeholder for selftest function. An example use would be to test package api connectivity.
    Suggested return values are be unimplemented, success, or failure.
    """

    return {
        "state": "success",
        "reason": None
    }


# -*- coding: utf-8 -*-
# (c) Copyright IBM Corp. 2010, 2018. All Rights Reserved.

"""Circuits component for Action Module subscription and message handling"""

import sys
import json
import re
import os.path
import random
import datetime
import logging
import traceback
import six
from circuits import Event, Timer

LOG = logging.getLogger(__name__)


class ActionMessageBase(Event):
    """Superclass for :class:`ActionMessage` and :class:`FunctionMessage`.
    """

    def __init__(self, source=None, headers=None, message=None,
                 test=False, test_msg_id=None, frame=None, log_dir=None):
        super(ActionMessageBase, self).__init__(source=source,
                                                headers=headers,
                                                message=message)
        if headers is None:
            headers = {}
        if message is None:
            message = {}
        LOG.debug("Source: %s", source)
        LOG.debug("Headers: %s", json.dumps(headers, indent=2))
        LOG.debug("Message: %s", json.dumps(message, indent=2))

        self.deferred = False
        self.message = message
        self.frame = frame
        self.context = headers.get("Co3ContextToken")
        self.action_id = message.get("action_id")
        self.object_type = message.get("object_type")
        self.test = test
        self.test_msg_id = test_msg_id

        self.timestamp = None
        ts = headers.get("timestamp")
        if ts is not None:
            self.timestamp = datetime.datetime.utcfromtimestamp(float(ts)/1000)

        self.name = "_unknown_"
        self.displayname = "Unknown"
        self.principal = (message.get("principal") or {}).get("name", "None")
        self.workflow = (message.get("workflow") or {}).get("programmatic_name", "None")

        # Fire a {name}_success event when this event is successfully processed
        self.success = True

    def __repr__(self):
        "x.__repr__() <==> repr(x)"
        if len(self.channels) > 1:
            channels = repr(self.channels)
        elif len(self.channels) == 1:
            channels = str(self.channels[0])
        else:
            channels = ""
        return "<%s[%s] (id=%s, workflow=%s, user=%s) %s>" % (self.name,
                                                              channels,
                                                              self.action_id,
                                                              self.workflow,
                                                              self.principal,
                                                              self.timestamp)

    def __getattr__(self, name):
        """Message attributes are made accessible as properties
           ("incident", "task", "note", "milestone". "task", "artifact";
           and "properties" for the action fields on manual actions)
        """
        if name == "message":
            raise AttributeError()
        try:
            return self.message[name]
        except KeyError:
            raise AttributeError()

    def hdr(self):
        """Get the headers (dict)"""
        return self.kwargs["headers"]

    def msg(self):
        """Get the message (dict)"""
        return self.kwargs["message"]

    def defer(self, component, delay=None):
        """Defer this message for handling later"""
        if self.deferred:
            # This message was already deferred.  You should just handle it.
            # (Mark it as no longer deferred, so that it will ack now)
            self.deferred = False
            return False
        # Fire me again after a dela
        if delay is None:
            delay = 0.5 + random.random()
        self.deferred = True
        LOG.debug("Deferring %s (%s)", self, self.hdr().get("message-id"))
        Timer(delay, self).register(component)
        return True

    def _log_message(self, log_dir):
        """Log Message JSON to File"""
        filename = "_".join((self.__class__.name, self.displayname,
                             datetime.datetime.now().isoformat())).replace('/', '_').replace(':', '-')
        with open(os.path.join(log_dir,
                               filename.format("JSON")), "w+") as logfile:
            logfile.write(json.dumps(self.message, indent=2))


class ActionMessage(ActionMessageBase):
    """A Circuits event for a Resilient Action Module message.

    This event holds details of the Action Module message,
    including its context (the incident, task, artifact... where the action
    was triggered).  Your components will receive these events from the
    Resilient Action Module message destination.

    These events are named by the rule that triggered them (lowercased).
    So a custom action rule named "Manual Action" will generate an event with name
    "manual_action".  To handle that event, you should implement a :class:`ResilientComponent`
    that has a method named :samp:`manual_action`:  the Circuits framework will call
    your component's methods based on the name of the event.

    The parameters for your event-handler method are:
       * event: this event object
       * source: the component that fired the event
       * headers: the Action Module message headers (dict)
       * message: the Action Module message (dict)
    For convenience, the message is also broken out onto event properties,
       * event.incident: the incident that the event relates to
       * event.artifact: the artifact that the event was triggered from (if any)
       * event.task: the task that the event was triggered from (if any)
         (etc).

    To have your component's method with a different name from the action,
    you can use the :func:`handler` decorator:

    .. code-block:: python

        @handler("the_action_name")
        def _any_method_name(self, event, source=None, headers=None, message=None):
            ...

    To have a method handle *any* event on the component's channel,
    use the :func:`handler` decorator with no event name,

    .. code-block:: python

        @handler()
        def _any_method_name(self, event, source=None, headers=None, message=None):
            ...

    Refer to the developer documentation for additional details on writing components and event handlers.
    """

    def __init__(self, source=None, headers=None, message=None,
                 test=False, test_msg_id=None, frame=None, log_dir=None):
        super(ActionMessage, self).__init__(source=source,
                                            headers=headers,
                                            message=message,
                                            test=test,
                                            test_msg_id=test_msg_id,
                                            frame=frame,
                                            log_dir=log_dir)

        self.action_id = message.get("action_id")

        if isinstance(source, str):
            # just for testing
            self.displayname = source
        elif source is not None:
            self.displayname = source.action_name(self.action_id)

        # The name of this event (=the function that subscribers implement)
        # is determined from the name of the action.
        # In future, this should be the action's "programmatic name",
        # but for now it's the downcased displayname with underscores.
        self.name = re.sub(r'\W+', '_', self.displayname.strip().lower())

        if message and log_dir:
            self._log_message(log_dir)


class FunctionMessage(ActionMessageBase):
    """A Circuits event for a Resilient Function call.

    This event holds details of the Function message, including its input parameters.
    Your components will receive these events from their Resilient message destination.

    To handle a function, you should implement a :class:`ResilientComponent` with a method
    annotated with the :func:`function` decorator:

    .. code-block:: python

        @function("the_function_name")
        def _any_method_name(self, event, *args, **kwargs):
            ...

    The parameters for your function method are:
       * event: this event object
       * args and kwargs: additional context
    The `event.message` contains:
       * user: details of the user who triggered the function call
       * function: details of the function being called
       * workflow: information about the workflow containing the function
       * inputs: the input parameter values.

    To return a value from your function, yield or return a :class:`FunctionResult` containing
    the value:

    .. code-block:: python

        yield FunctionResult("xyz")

    Refer to the developer documentation for additional details on writing components and event handlers.
    """

    def __init__(self, source=None, headers=None, message=None,
                 test=False, test_msg_id=None, frame=None, log_dir=None):
        super(FunctionMessage, self).__init__(source=source,
                                              headers=headers,
                                              message=message,
                                              test=test,
                                              test_msg_id=test_msg_id,
                                              frame=frame,
                                              log_dir=log_dir)

        fn = message["function"]
        self.action_id = fn.get("id")

        # The name of this event is the API name name of the function.
        self.name = fn.get("name", "_unnamed_")
        self.displayname = fn.get("display_name", self.name)

        if message and log_dir:
            self._log_message(log_dir)


class InboundMessage(ActionMessageBase):
    def __init__(self, source=None, headers=None, message=None, queue=None,
                 test=False, test_msg_id=None, frame=None, log_dir=None):

        super(InboundMessage, self).__init__(source=source, headers=headers, message=message,
                                             test=test, test_msg_id=test_msg_id, frame=frame, log_dir=log_dir)

        assert isinstance(queue, tuple)
        assert len(queue) == 3

        self.name = queue[2]
        self.displayname = queue[2]

        if message and log_dir:
            self._log_message(log_dir)


class StatusMessage(object):
    """Encapsulates a status message yielded from an action or function call"""
    def __init__(self, text):
        super(StatusMessage, self).__init__()
        self.text = text

    def __str__(self):
        return self.text or ""


class FunctionResult(object):
    def __init__(self, value, success=True, reason=None, name="Unknown", custom_results=False):
        """
        Encapsulates the result of a Function call

        :param value: the results of the Function
        :type value: dict
        :param success: whether or not the Function has succeeded, defaults to ``True``
        :type success: bool, optional
        :param reason: a brief reason why it failed, defaults to ``None``
        :type reason: str | None, optional
        :param name: name of the Function, defaults to ``"Unknown"``
        :type name: str, optional
        :param custom_results: if ``True`` ``value`` will not contain any default payload.
            We **do not recommend setting it to True** however in some cases it may be necessary,
            defaults to ``False``
        :type custom_results: bool, optional
        """
        super(FunctionResult, self).__init__()
        if not isinstance(value, dict):
            msg = "FunctionResult must be a dictionary. " \
                  "'{}' may cause the workflow to fail.".format(type(value).__name__)
            logging.getLogger(__name__).warning(msg)

        self.value = value
        self.success = success
        self.reason = reason
        self.name = name
        self.custom_results = custom_results


def FunctionError(*args, **kwargs):
    """
    Way to pass an error message back to the client.
    Usage in the function:
    yield FunctionError("Message goes here")

    if an exception happened prior to this, if will return FunctionException_,
    otherwise it will return FunctionError_. Both of them are subclassing BaseFunctionError, which
    makes it easier to handle Function Errors specifically.

    :param: trace - boolean that shows whether the trace should be passed on
    """
    # can't just include it into params, because Python2.7 doesn't allow args to be first
    # and doing it otherwise might break backwards compatibility
    trace = kwargs.pop("trace", True)

    # Just grab the stack trace and wrap in a FunctionError_.
    exc = sys.exc_info()[0]
    exc_trace = traceback.format_exc()
    if not exc:
        return FunctionError_(*args)
    return FunctionException_(*args, include_trace=trace, trace=exc_trace)


class BaseFunctionError(ValueError):
    """
    Provides an extra layer between FunctionError_/FunctionException_ that allows to check
    more precisely if it's a resilient-circuits error, and not a ValueError, but doesn't require
    to check for both separately.
    """

    def __init__(self, *args, **kwargs):
        include_trace = kwargs.pop("include_trace", False)
        trace = kwargs.pop("trace", None)

        super(BaseFunctionError, self).__init__(*args, **kwargs)
        self.include_trace = include_trace
        self.trace = trace

    def __str__(self):
        try:
            message = unicode(self).encode('utf-8')
        except NameError:
            # not Python 2
            message = self.__unicode__()
        return message

    def __unicode__(self):
        """
        Gets called by unicode in Python 2.
        :return: unicode representation of the class
        """
        message = u""
        if isinstance(self.args, (tuple, list)) and len(self.args) > 0:
            if isinstance(self.args[0], six.string_types):
                try:
                    message += self.args[0]
                except UnicodeDecodeError:
                    # already unicode in Py2
                    message = message + self.args[0].decode('utf-8')
        if self.include_trace and self.trace is not None:
            message += u"\n"
            try:
                message += self.trace
            except UnicodeDecodeError:
                # already unicode in Py2
                message = message + self.trace.decode('utf-8')
        return message

class FunctionException_(BaseFunctionError):
    """Wraps an exception from a function call."""
    pass

class FunctionError_(BaseFunctionError):
    """Wraps a simple "we failed" error from a function call."""
    pass

class StatusMessageEvent(Event):
    """Event that we use to send "action status" update back to resilient"""
    def __init__(self, parent=None, message=None):
        super(StatusMessageEvent, self).__init__(message)
        self.parent = parent

    @property
    def text(self):
        """Text of the message"""
        if isinstance(self.args, (tuple, list)) and len(self.args) > 0:
            return self.args[0]
        return ""


class FunctionErrorEvent(Event):
    """Event that we use to send "action failure" update back to resilient"""
    def __init__(self, parent=None, message=None):
        super(FunctionErrorEvent, self).__init__(message)
        self.parent = parent

    @property
    def text(self):
        """Text of the message"""
        if isinstance(self.args, (tuple, list)) and len(self.args) > 0:
            return self.args[0]
        return ""



# -*- coding: utf-8 -*-
# (c) Copyright IBM Corp. 2010, 2018. All Rights Reserved.

"""Dynamic component loader"""

import os
import sys
import logging
import traceback
import pkg_resources
from circuits import Loader, Event
from circuits.core.handlers import handler
from resilient_circuits import constants, helpers

LOG = logging.getLogger(__name__)


class load(Event):
    """yes"""
    complete = True


class load_all_failure(Event):
    """Event indicating that a component failed to load."""
    pass


class load_all_success(Event):
    """Event indicating that all components were loaded."""
    pass


def safe_but_noisy_import(name):
    modules = sys.modules.copy()
    try:
        if name in sys.modules:
            LOG.debug("Name exists in modules")
            return reload(sys.modules[name])
        else:
            LOG.debug("Name does not exist in modules")
            return __import__(name, globals(), locals(), [""])
    except Exception as exc:
        for name in sys.modules.copy():
            if name not in modules:
                del sys.modules[name]
        LOG.exception(exc)


class ComponentLoader(Loader):
    """A component to automatically load from the componentsdir directory"""

    def __init__(self, opts):
        """Initialize the loader"""
        self.opts = opts
        # Path where components should be found
        self.path = opts['componentsdir']
        # Optionally, a list of filenames that should not be loaded
        noload = opts.get("noload", "")
        self.noload = [filename.strip() for filename in noload.split(",") if filename.strip() != ""]
        super(ComponentLoader, self).__init__(init_kwargs={"opts": opts}, paths=[self.path])
        self.pending_components = []
        self.finished = False

        # Load all installed components
        installed_components = self.discover_installed_components()
        if installed_components:
            self._register_components(installed_components)

        if self.path:
            # Load all components from the components directory
            for filename in os.listdir(self.path):
                filepath = os.path.join(self.path, filename)
                if os.path.isfile(filepath) and os.path.splitext(filename)[1] == ".py":
                    cname = os.path.splitext(filename)[0]
                    if cname != "__init__":
                        if cname in self.noload:
                            LOG.info("Not loading %s", cname)
                        else:
                            LOG.info("Loading '%s' from %s", cname, filepath)
                            self.pending_components.append(cname)
                            self.fire(load(cname))
        if not self.pending_components:
            # No components from directory, we are done loading
            self.finished = True
            self.fire(load_all_success())

    def discover_installed_components(self):
        entry_points = pkg_resources.iter_entry_points('resilient.circuits.components')
        ep = None
        try:
            return_list = []
            # Loop entry points
            for ep in entry_points:
                if ep.name not in self.noload:

                    # Load the component class
                    cmp_class = ep.load()

                    # Get the class' __module__ name
                    # Note: the name used for the app.config section of this app should be the same as the __module__
                    cmp_module_name = cmp_class.__module__.split(".")[0]

                    # If an INBOUND_MSG_APP_CONFIG_Q_NAME is defined in the app.config in the section for this app, use it
                    custom_q_name = self.opts.get(cmp_module_name, {}).get(constants.INBOUND_MSG_APP_CONFIG_Q_NAME, "")

                    if custom_q_name:
                        # Get the inbound_handlers in this component and overwite their 'channel' and 'names' attributes
                        inbound_handlers = helpers.get_handlers(cmp_class, handler_type="inbound_handler")
                        for ih in inbound_handlers:

                            new_channel = "{0}.{1}".format(constants.INBOUND_MSG_DEST_PREFIX, custom_q_name)
                            new_names = (custom_q_name, )

                            if sys.version_info.major < 3:
                                # Handle PY < 3
                                ih[1].__func__.channel = new_channel
                                ih[1].__func__.names = new_names
                            else:
                                # Handle PY >= 3 specific imports
                                ih[1].channel = new_channel
                                ih[1].names = new_names

                    return_list.append(cmp_class)
            return return_list

        except ImportError as e:
            LOG.error("Failed to load '%s' from '%s'", ep, ep.dist)
            raise e

    def _register_components(self, component_list):
        """ register all installed components and ones from componentsdir """
        LOG.info("Loading %d components", len(component_list))
        for component_class in component_list:
            LOG.info("'%s.%s' loading", component_class.__module__, component_class.__name__)
            try:
                component_class(opts=self.opts).register(self)
                LOG.debug("'%s.%s' loaded", component_class.__module__, component_class.__name__)
            except Exception as e:
                LOG.error("Failed to load '%s.%s'", component_class.__module__, component_class.__name__, exc_info=1)
                self.fire(load_all_failure())
                return False
        return True

    @handler("exception", channel="loader")
    def exception(self, event, *args, **kwargs):
        if not self.finished:
            fevent = kwargs.get("fevent", None)
            if fevent is not None:
                cname = fevent.args[0]
                LOG.error("An exception occurred while loading component '%s'", cname)
                self.fire(load_all_failure())

    @handler("load_complete")
    def load_complete(self, event, *eargs, **ekwargs):
        """Check whether the component loaded successfully"""
        cname = event.args[0]
        if isinstance(cname, Event):
            cname = cname.args[0]

        if cname in sys.modules:
            LOG.info("Loaded and registered component '%s'", cname)
            self.pending_components.remove(cname)
            if self.pending_components == []:
                self.finished = True
                self.fire(load_all_success())
        else:
            LOG.error("Failed to load component '%s'", cname)
            safe_but_noisy_import(cname)
            self.fire(load_all_failure())



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# (c) Copyright IBM Corp. 2010, 2018. All Rights Reserved.

"""Argument parse helper, for values stored in keyring or file"""

import resilient


class ArgumentParser(resilient.ArgumentParser):
    """Nothing here.  All this functionality is now in the base class"""



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# (c) Copyright IBM Corp. 2010, 2020. All Rights Reserved.

"""Common Helper Functions for resilient-circuits"""
import sys
import pkg_resources
import logging
import copy
import re
import time
from resilient_circuits import constants
from resilient import get_client
from resilient import is_env_proxies_set, get_and_parse_proxy_env_var
from resilient import constants as res_constants


LOG = logging.getLogger("__name__")


def get_fn_names(component):
    """If `component` has a `function` attribute and it is True,
    appends the names in the function handler to `fn_names` and
    returns it, else returns an empty list.

    :param component: the component object to get it's list of function names for
    :type component: object
    :return: fn_names: the name in each function handler in the component if found
    :rtype: list
    """

    assert isinstance(component, object)

    fn_names = []

    # Get a list of callable methods for this object
    methods = [a for a in dir(component) if callable(getattr(component, a))]

    for m in methods:
        this_method = getattr(component, m)
        is_function = getattr(this_method, "function", False)

        if is_function:
            fn_decorator_names = this_method.names
            # Fail if fn_decorator_names is not a tuple as may have unhandled side effects if a str etc.
            # When a function handler is decorated its __init__() function takes the '*args' parameter
            # When * is prepended, it is known as an unpacking operator to allow the function handler to have
            # multiple names. args (or names in our case) will be a tuple, so if the logic of the function
            # decorator changes, this will catch it.
            assert isinstance(fn_decorator_names, tuple)
            for n in fn_decorator_names:
                fn_names.append(n)

    return fn_names


def get_handlers(component, handler_type="inbound_handler"):
    """If `component` has a `handler_type` attribute and it is True,
    appends a tuple to the handlers list and returns the list,
    else returns an empty list.

    Return example:
        -  `[(<method_name>, <method>, <handler_type>)]`
        -  `[('_inbound_app_mock_one', <function InboundAppComponent._inbound_app_mock_one at 0x10ccc9510>, 'inbound_handler')]`

    :param component: the component object to check if it's methods is a `handler_type`
    :type component: object
    :return: handlers: the name in each function handler in the component if found
    :rtype: list of tuples
    """

    assert isinstance(component, object)
    assert isinstance(handler_type, str)

    handlers = []

    # Get a list of callable methods for this object
    methods = [a for a in dir(component) if callable(getattr(component, a))]

    for m in methods:
        this_method = getattr(component, m)
        is_handler = getattr(this_method, handler_type, False)

        if is_handler:
            handlers.append((m, this_method, handler_type))

    return handlers


def check_exists(key, dict_to_check):
    """Returns the value of the key in dict_to_check if found,
    else returns False. If dict_to_check is None, returns False

    :param key: the key to look for in dict_to_check
    :type key: str
    :param dict_to_check: the key to look for in dict_to_check
    :type dict_to_check: dict
    :return: value of key in dict_to_check else False
    """
    if dict_to_check is None:
        return False

    assert isinstance(dict_to_check, dict)

    return dict_to_check.get(key, False)


def get_configs(path_config_file=None, ALLOW_UNRECOGNIZED=False):
    """
    Gets all the configs that are defined in the app.config file
    Uses the path to the config file from the parameter
    Or uses the `get_config_file()` method in resilient if None

    :param path_config_file: path to the app.config to parse
    :type path_config_file: str
    :param ALLOW_UNRECOGNIZED: bool to specify if AppArgumentParser will allow unknown comandline args or not. Default is False
    :type ALLOW_UNRECOGNIZED: bool
    :return: dictionary of all the configs in the app.config file
    :rtype: dict
    """
    from resilient import get_config_file
    from resilient_circuits.app_argument_parser import AppArgumentParser

    if not path_config_file:
        path_config_file = get_config_file()

    configs = AppArgumentParser(config_file=path_config_file).parse_args(ALLOW_UNRECOGNIZED=ALLOW_UNRECOGNIZED)
    return configs


def get_resilient_client(path_config_file=None, ALLOW_UNRECOGNIZED=False):
    """
    Return a SimpleClient for Resilient REST API using configurations
    options from provided path_config_file or from ~/.resilient/app.config

    :param path_config_file: path to the app.config to parse
    :type path_config_file: str
    :param ALLOW_UNRECOGNIZED: bool to specify if AppArgumentParser will allow unknown comandline args or not. Default is False
    :type ALLOW_UNRECOGNIZED: bool
    :return: SimpleClient for Resilient REST API
    :rtype: SimpleClient
    """
    client = get_client(get_configs(path_config_file=path_config_file, ALLOW_UNRECOGNIZED=ALLOW_UNRECOGNIZED))
    return client


def validate_configs(configs, validate_dict):
    """
    Checks if the configs are valid and raise a ValueError if they are not.
    Check if the config is required, has a value and meets its 'condition'

    :param configs: normally the configs in app.config
    :type configs: dict
    :param validate_dict: the key is the config and the value is a dict with the following params:
        - **required**: a boolean if the config is required
        - **placeholder_value**: the default value of the config that is not valid
        - **valid_condition**: a function to check if the value of the config is valid e.g. within a certain range
        - **invalid_msg**: displayed when valid_condition fails
    :type validate_dict: dict
    :return: nothing
    """
    if not isinstance(configs, dict):
        raise ValueError("'configs' must be of type dict, not {0}".format(type(configs)))

    if not isinstance(validate_dict, dict):
        raise ValueError("'validate_dict' must be of type dict, not {0}".format(type(validate_dict)))

    for config_name, config_validations in validate_dict.items():

        required = config_validations.get("required")
        placeholder_value = config_validations.get("placeholder_value")
        valid_condition = config_validations.get("valid_condition", lambda c: True)
        invalid_msg = config_validations.get("invalid_msg", "'{0}' did not pass it's validate condition".format(config_name))

        # get the config value from configs
        config = configs.get(config_name)

        # if its required
        if required:

            # if not in configs or empty string
            if not config:
                raise ValueError("'{0}' is mandatory and is not set in the config file.".format(config_name))

        # if still equals placeholder value
        if placeholder_value and config == placeholder_value:
            raise ValueError("'{0}' is mandatory and still has its placeholder value of '{1}' in the config file.".format(config_name, placeholder_value))

        # if meets its valid_condition
        if not valid_condition(config):
            raise ValueError(invalid_msg)


def get_packages(working_set):
    """
    Return a sorted list of tuples of all package names
    and their version in working_set

    :param working_set: the working_set for all packages installed in this env
    :type working_set: setuptools.pkg_resources.WorkingSet obj
    :return: pkg_list: a list of tuples [('name','version')] e.g. [('resilient-circuits', '39.0.0')].
        If ``working_set`` is not a ``pkg_resources.WorkingSet`` object, just return the current value of ``working_set``
    :rtype: list
    """

    if not isinstance(working_set, pkg_resources.WorkingSet):
        return working_set

    pkg_list = []

    for pkg in working_set:
        pkg_list.append((pkg.project_name, pkg.version))

    return sorted(pkg_list, key=lambda x: x[0].lower())


def get_env_str(packages):
    """
    Return a str with the Python version and the
    packages

    :param packages: the working_set for all packages installed in this env
    :type packages: setuptools.pkg_resources.WorkingSet obj
    :return: env_str: a str of the Environment
    :rtype: str
    """

    env_str = u"{0}Environment:\n".format(constants.LOG_DIVIDER)
    env_str += u"Python Version: {0}\n\n".format(sys.version)
    env_str += u"Installed packages:\n"
    for pkg in get_packages(packages):
        env_str += u"\n\t{0}: {1}".format(pkg[0], pkg[1])

    if is_env_proxies_set():

        proxy_details = get_and_parse_proxy_env_var(res_constants.ENV_HTTPS_PROXY)

        if not proxy_details:
            proxy_details = get_and_parse_proxy_env_var(res_constants.ENV_HTTP_PROXY)

        if proxy_details:
            env_str += u"\n\nConnecting through proxy: '{0}://{1}:{2}'".format(proxy_details.get("scheme"), proxy_details.get("hostname"), proxy_details.get("port"))

    env_str += u"\n###############"
    return env_str




def get_queue(destination):
    """
    Return a tuple (queue_type, org_id, queue_name).
    Returns None if fails to get queue

    :param destination: str in the format '/queue/actions.201.fn_main_mock_integration'
    :type destination: str
    :return: queue: (queue_type, org_id, queue_name) e.g. ('actions', '201', 'fn_main_mock_integration')
    :rtype: tuple
    """

    try:
        assert isinstance(destination, str)

        regex = re.compile(r'\/.+\/')

        destination = re.sub(regex, "", destination, count=1)
        q = destination.split(".")

        assert len(q) == 3

        return (tuple(q))

    except AssertionError as e:
        LOG.error("Could not get queue name\n%s", str(e))
        return None


def is_this_a_selftest(component):
    """
    Return ``True`` or ``False`` if this instantiation of
    ``resilient-circuits`` is from selftest or not.

    :param component: the current component that is calling this method (usually ``self``)
    :type component: circuits.Component
    :rtype: bool
    """
    return component.IS_SELFTEST


def should_timeout(start_time, timeout_value):
    """
    Returns True if the delta between the
    start_time and the current_time is greater

    All time values are the time in seconds since
    the epoch as a floating point number

    :param start_time: the time before the loop starts
    :type start_time: float
    :param timeout_value: number of seconds to timeout after
    :type timeout_value: int/float
    :rtype: bool
    """
    return (time.time() - start_time) > timeout_value


def get_user(app_configs):
    """
    Looks for either 'api_key_id' or 'email'
    in the provided dict and returns its value.

    Returns None if neither are found or set to
    a valid value

    :param app_configs: a dictionary of all the values in the app.config file
    :type app_configs: dict
    :rtype: [str/None]
    """
    usr = app_configs.get("api_key_id", None)

    if not usr:
        usr = app_configs.get("email", None)

        if not usr:
            return None

    return usr


def filter_heartbeat_timeout_events(heartbeat_timeouts):
    """
    Sort a list of HeartbeatTimeout on their ts and
    remove an occurrences of HeartbeatTimeout whose ts
    value is -1

    :param heartbeat_timeouts: List of HeartbeatTimeout
    :type heartbeat_timeouts: [HeartbeatTimeout]
    :return: Sorted and filtered list of HeartbeatTimeout
    :rtype: [HeartbeatTimeout]
    """

    if not heartbeat_timeouts:
        return heartbeat_timeouts

    # Remove any elements whose ts is -1
    heartbeat_timeouts = [hb for hb in heartbeat_timeouts if hb.ts != -1]

    heartbeat_timeouts.sort()

    return heartbeat_timeouts



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# (c) Copyright IBM Corp. 2010, 2023. All Rights Reserved.

import pytest
from mock import MagicMock, patch

from resilient_circuits.stomp_component import StompClient, StompConnectionError
from resilient_circuits.stomp_events import ConnectionFailed
from resilient_circuits import constants

def test_stomp_reconnect_two_consecutive_failures(caplog):

    client = StompClient(host="example.com", port=443)

    assert client._stomp_max_connection_errors == constants.STOMP_MAX_CONNECTION_ERRORS

    # standard successfull connection
    mock_evt = MagicMock()
    with patch("resilient_circuits.stomp_component.Stomp.connect") as patch_stomp_library_connect:
        with patch("resilient_circuits.stomp_component.StompClient.connected") as patch_connected_state:
            patch_stomp_library_connect.side_effect = None
            patch_connected_state.return_value = True
            client.connect(mock_evt)


            # first a StompConnectionError with a different error (this could be possible in real world)
            patch_stomp_library_connect.side_effect = StompConnectionError("A different STOMP error")
            client.connect(mock_evt)
            assert client._stomp_connection_errors == 0

            # now a StompConnectionError with "no more data" error
            # this should trigger the max, and throw a sys.exit(1)
            patch_stomp_library_connect.side_effect = StompConnectionError("Some preamble. No more data received from STOMP")
            with pytest.raises(SystemExit):
                client.connect(mock_evt)
                assert "Exiting due to unrecoverable error" in caplog.text
            assert client._stomp_connection_errors == 1

def test_stomp_reconnect_one_failure_followed_by_successful_reconnect(caplog):

    max_connection_errors_override = 5
    client = StompClient(host="example.com", port=443, stomp_max_connection_errors=max_connection_errors_override)

    assert client._stomp_max_connection_errors == max_connection_errors_override

    # standard successfull connection
    mock_evt = MagicMock()
    with patch("resilient_circuits.stomp_component.Stomp.connect") as patch_stomp_library_connect:
        with patch("resilient_circuits.stomp_component.StompClient.connected") as patch_connected_state:
            patch_stomp_library_connect.side_effect = None
            patch_connected_state.return_value = True
            client.connect(mock_evt)

            # as we've changed the max_connection errors limit to 5, run this 4 times but
            # then reconnect successfully and observe it is reset to 0
            for i in range(max_connection_errors_override-1):
                # first StompConnectionError with "no more data" error
                patch_stomp_library_connect.side_effect = StompConnectionError("No more data received from STOMP")
                client.connect(mock_evt)
                assert client._stomp_connection_errors == i+1

            # successful reconnect now should reset count to 0
            patch_stomp_library_connect.side_effect = None
            client.connect(mock_evt)
            assert client._stomp_connection_errors == 0


#!/usr/bin/env python
# -*- coding: utf-8 -*-
# (c) Copyright IBM Corp. 2010, 2021. All Rights Reserved.

"""
Common place for helper functions used in tests
"""

from resilient_circuits import SubmitTestInboundApp, SubmitTestFunction, FunctionResult


def call_app_function(fn_name, fn_inputs, circuits_app, status_message_only=False, full_result_obj=False):

    assert isinstance(fn_inputs, dict)

    # Create the submitTestFunction event
    evt = SubmitTestFunction(fn_name, fn_inputs)

    # Fire a message to the function
    circuits_app.manager.fire(evt)

    # circuits will fire an "exception" event if an exception is raised in the ResilientComponent
    # return this exception if it is raised
    exception_event = circuits_app.watcher.wait("exception", parent=None, timeout=2)

    if exception_event:
        exception = exception_event.args[1]
        raise exception

    if status_message_only:
        status_message_event = circuits_app.watcher.wait("StatusMessageEvent", parent=None, timeout=2)
        return status_message_event.value.event

    event = circuits_app.watcher.wait(fn_name + u"_result", parent=evt, timeout=2)
    result = event.kwargs.get("result")
    assert isinstance(result, FunctionResult)

    if full_result_obj:
        return result

    return result.value


def call_inbound_app(circuits_app, queue_name, action="create", msg_content={}, message=None):
# TODO: add doc string
    evt = SubmitTestInboundApp(queue_name, action, msg_content, message)

    # Fire a message to the inbound app
    circuits_app.manager.fire(evt)

    # circuits will fire an "exception" event if an exception is raised in the ResilientComponent
    # return this exception if it is raised
    exception_event = circuits_app.watcher.wait("exception", parent=None, timeout=2)

    if exception_event:
        exception = exception_event.args[1]
        raise exception

    event = circuits_app.watcher.wait(queue_name + u"_success", parent=evt, timeout=2)

    if event:
        return event.args

    return False



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# (c) Copyright IBM Corp. 2010, 2018. All Rights Reserved.

"""Service wrapper.  Runs the main python application from a Windows service."""

import win32service
import win32serviceutil
import win32api
import win32con
import win32job
import subprocess
import os
import signal
import shlex

SERVICE_NAME = "RESIL_SVC"
SERVICE_DISPLAY_NAME = "Resilient Circuits"


class irms_svc(win32serviceutil.ServiceFramework):
    """The service implementation"""

    _svc_name_ = SERVICE_NAME
    _svc_display_name_ = SERVICE_DISPLAY_NAME
    process_handle = None
    _resilient_args_ = ""

    def __init__(self, args):
        win32serviceutil.ServiceFramework.__init__(self, args)
        self.isAlive = True

    @classmethod
    def setResilientArgs(cls, arg_string):
        """args to pass to resilient-circuits.exe run.  Feature does not work yet"""
        cls._resilient_args_ = arg_string
        # TODO: Store these somwhere permanent (registry?) and the reload them when running

    def SvcDoRun(self):
        import servicemanager
        servicemanager.LogInfoMsg(self._svc_name_ + " Start Requested")
        try:
            hJob = win32job.CreateJobObject(None, "")
            extended_info = win32job.QueryInformationJobObject(hJob, win32job.JobObjectExtendedLimitInformation)
            extended_info['BasicLimitInformation']['LimitFlags'] = win32job.JOB_OBJECT_LIMIT_KILL_ON_JOB_CLOSE
            win32job.SetInformationJobObject(hJob, win32job.JobObjectExtendedLimitInformation, extended_info)
            command = "resilient-circuits.exe run " + self._resilient_args_
            command_args = shlex.split(command)
            self.process_handle = subprocess.Popen(command_args)
            # Convert process id to process handle:
            perms = win32con.PROCESS_TERMINATE | win32con.PROCESS_SET_QUOTA
            hProcess = win32api.OpenProcess(perms, False, self.process_handle.pid)
            win32job.AssignProcessToJobObject(hJob, hProcess)
        except:
            servicemanager.LogErrorMsg(self._svc_name_ + " failed to launch resilient-circuits.exe")
            raise
        servicemanager.LogInfoMsg(self._svc_name_ + " Started")

        while self.isAlive:
            if self.process_handle.poll() != None:
                self.SvcStop()
            win32api.SleepEx(10000, True)

    def SvcStop(self):
        import servicemanager
        msg = "stopping."
        try:
            # There's no way to say this nicely
            self.process_handle.terminate()
        except Exception as exc:
            msg = exc.message
            pass
        servicemanager.LogInfoMsg(self._svc_name_ + " - Received stop signal; " + msg)
        self.ReportServiceStatus(win32service.SERVICE_STOP_PENDING)
        self.isAlive = False


def ctrlHandler(ctrlType):
    return True


if __name__ == '__main__':
    win32api.SetConsoleCtrlHandler(ctrlHandler, True)
    win32serviceutil.HandleCommandLine(irms_svc)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# (c) Copyright IBM Corp. 2010, 2021. All Rights Reserved.

import pytest
from resilient_circuits.action_message import ActionMessageBase, InboundMessage, FunctionResult
from tests import mock_constants


class TestFunctionResult:

    def test_basic_instantiation(self):

        mock_results = {"content": u"unicode: ล ฦ ว"}
        mock_fn_name = "mock_fn_name"

        result = FunctionResult(mock_results, name=mock_fn_name)

        assert result.value == mock_results
        assert result.name == mock_fn_name
        assert result.reason is None
        assert result.success is True
        assert result.custom_results is False

    def test_custom_result_instantiation(self):

        mock_results = {"content": u"unicode: ล ฦ ว"}
        mock_fn_name = "mock_fn_name"

        result = FunctionResult(mock_results, name=mock_fn_name, success=False, reason="Failure", custom_results=True)

        assert result.value == mock_results
        assert result.name == mock_fn_name
        assert result.reason == "Failure"
        assert result.success is False
        assert result.custom_results is True

    def test_results_not_dict(self, caplog):

        error_msg = "FunctionResult must be a dictionary"
        FunctionResult("")
        assert error_msg in caplog.text


class TestInboundMessage:

    def test_basic_instantiation(self):

        mock_event = InboundMessage(queue=mock_constants.MOCK_QUEUE)
        assert isinstance(mock_event, ActionMessageBase)
        assert isinstance(mock_event, InboundMessage)
        assert mock_event.name == mock_constants.MOCK_QUEUE_NAME

    def test_queue_not_tuple(self):
        with pytest.raises(AssertionError):
            InboundMessage(queue="no queue")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# (c) Copyright IBM Corp. 2010, 2021. All Rights Reserved.

"""Implementation of AppFunctionComponent"""

import logging
import threading
from collections import namedtuple
from resilient_circuits import ResilientComponent, handler, StatusMessage
from resilient_lib import RequestsCommon, validate_fields


class AppFunctionComponent(ResilientComponent):
    """
    Each package that has been generated with the ``resilient-sdk codegen`` command
    that contains a SOAR Function, an :class:`AppFunctionComponent` will be generated
    in the package's ``components`` directory

    Each :class:`AppFunctionComponent`, gets loaded into the ``resilient-circuits`` framework
    and listens for messages on it's ``Message Destination``
    """

    def __init__(self, opts, package_name, required_app_configs=[]):
        """
        Sets and exposes 4 properties:
            #. **self.PACKAGE_NAME**: the name of this App, the parameter passed in
            #. **self.app_configs**: a `collections.namedtuple <https://docs.python.org/3.6/library/collections.html#collections.namedtuple>`_
               object of all associated configurations in the ``app.config`` file for this. App Configurations can be accessed like:

               .. code-block:: python

                     base_url = self.app_configs.base_url

            #. **self.rc**: an instantiation of :class:`resilient_lib.RequestsCommon <resilient_lib.components.requests_common.RequestsCommon>`
               which will have access to the ``execute`` method for external API calls
            #. **self.LOG**: access to a common ``Logger`` object. Can be used like:

               .. code-block:: python

                     self.LOG.info("Starting '%s'", self.PACKAGE_NAME)

        :param opts: all configurations from the ``app.config`` file
        :type opts: dict
        :param package_name: the name of this App
        :type package_name: str
        :param required_app_configs: a list of required ``app.config`` configurations
            that are required to run this App
        :type required_app_configs: [str, str, ...]
        :raises ValueError: if a ``required_app_config`` is not found in ``opts``
        """

        self.PACKAGE_NAME = package_name

        self._required_app_configs = required_app_configs

        # Validate app_configs and get dictionary as result
        self._app_configs_as_dict = validate_fields(required_app_configs, opts.get(package_name, {}))

        # This variable also is used to get the app.configs
        self.options = self._app_configs_as_dict

        # Instansiate RequestsCommon with dictionary of _app_configs_as_dict
        self.rc = RequestsCommon(opts=opts, function_opts=self._app_configs_as_dict)

        # Convert _app_configs_as_dict to namedtuple
        self.app_configs = namedtuple("app_configs", self._app_configs_as_dict.keys())(*self._app_configs_as_dict.values())

        self._local_storage = threading.local()

        self.LOG = logging.getLogger(__name__)

        super(AppFunctionComponent, self).__init__(opts)

    @property
    def required_app_configs(self):
        return self._required_app_configs

    @required_app_configs.setter
    def required_app_configs(self, value):
        self._required_app_configs = value

    @handler("reload")
    def _reload(self, event, opts):
        self.app_configs = validate_fields(self.required_app_configs, opts.get(self.PACKAGE_NAME, {}))

    @staticmethod
    def status_message(message):
        """
        Returns the message encapsulated in a
        ``resilient_circuits.StatusMessage`` object

        **Usage:**

        .. code-block:: python

            yield self.status_message("Finished running App Function: '{0}'".format(FN_NAME))

        :param message: Message you want to send to Action Status
        :type message: str

        :return: message encapsulated as ``StatusMessage``
        :rtype: ``resilient_circuits.StatusMessage``
        """
        return StatusMessage(message)

    def set_fn_msg(self, message_dict):
        """
        Uses threading.local() to store the message received
        locally for this Thread. Is accessed using the
        `get_fn_msg()` method below

        :param message_dict: Message received from SOAR
        :type message: dict
        """
        self._local_storage.fn_msg = message_dict

    def get_fn_msg(self):
        """
        Get the STOMP Message that has been sent from
        SOAR as a dict. The contents of the Message
        can include:

        .. code-block:: python

            'function': {
                'creator': '',
                'description': '',
                'display_name': 'mock_function',
                'id': 3,
                'name': 'mock_function',
                'tags': [...],
                'uuid': '',
                'version': None,
                'view_items': [...],
                ...
            },
            'groups': [],
            'inputs': {
                'mock_input_one': True,
                'mock_input_two': 'abc',
            },
            'playbook_instance': None,
            'principal': {
                'display_name': '',
                'id': 1,
                'name': '',
                'type': 'user'
            },
            'workflow': {
                'actions': [...],
                'description': '',
                'name': '',
                'object_type': { ... },
                'programmatic_name': '',
                'tags': [...],
                'uuid': '',
                'workflow_id': 3
            },
            'workflow_instance': {
                'workflow_instance_id': 54
            }

        :return: STOMP Message received from SOAR
        :rtype: dict
        """
        fn_msg = {}

        try:
            fn_msg = self._local_storage.fn_msg
        except AttributeError as err:
            self.LOG.warning("fn_msg could not be found\n%s", err)

        return fn_msg
